# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import io
import os
import unittest

import pytest

import virtinst

from tests import utils


# Misc tests for various virtinst special behavior. These should only aim for
# testing any particularly tricky bits, general XML generation should
# be through virt-install examples in test_cli


def test_misc_cpu_topology():
    """
    Various topology calculation corner cases
    """
    conn = utils.URIs.open_testdefault_cached()

    def get_top(_c):
        return [_c.topology.sockets, _c.topology.dies, _c.topology.cores, _c.topology.threads]

    cpu = virtinst.DomainCpu(conn)
    cpu.set_topology_defaults(6)
    assert cpu.topology.sockets is None

    cpu = virtinst.DomainCpu(conn)
    cpu.set_topology_defaults(6, create=True)
    assert get_top(cpu) == [1, 1, 6, 1]

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.sockets = "2"
    cpu.set_topology_defaults(6)
    assert get_top(cpu) == [2, 1, 3, 1]

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.dies = "3"
    cpu.set_topology_defaults(9)
    assert get_top(cpu) == [1, 3, 3, 1]

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.cores = "4"
    cpu.set_topology_defaults(8)
    assert get_top(cpu) == [2, 1, 4, 1]

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.threads = "3"
    cpu.set_topology_defaults(12)
    assert get_top(cpu) == [1, 1, 4, 3]

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.threads = "3"
    with pytest.raises(ValueError):
        cpu.set_topology_defaults(14)

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.sockets = 5
    cpu.topology.cores = 2
    assert cpu.vcpus_from_topology() == 10

    cpu = virtinst.DomainCpu(conn)
    cpu.topology.sockets = 3
    cpu.topology.dies = 2
    cpu.topology.cores = 2
    assert cpu.vcpus_from_topology() == 12

    cpu = virtinst.DomainCpu(conn)
    assert cpu.vcpus_from_topology() == 1


def test_misc_guest_osinfo_metadata():
    """
    Test that reading an unknown OS ID from guest XML will not blow up
    """
    # pylint: disable=protected-access
    g = virtinst.Guest(utils.URIs.open_testdefault_cached())
    g._metadata.libosinfo.os_id = "http://fedoraproject.org/fedora/20"
    assert g.osinfo.name == "fedora20"

    g = virtinst.Guest(utils.URIs.open_testdefault_cached())
    g._metadata.libosinfo.os_id = "http://example.com/idontexit"
    assert g.osinfo.name == "generic"


def test_misc_nonpredicatble_generate():
    """
    Bypass our testsuite 'predicatable' handling to test actual random output
    """
    from virtinst import cli

    predconn = utils.URIs.open_testdefault_cached()
    kvm_uri = utils.URIs.kvm_x86.replace(",predictable", "")
    kvmconn = cli.getConnection(kvm_uri)
    testconn = cli.getConnection("test:///default")

    testuuid = virtinst.Guest.generate_uuid(predconn)
    randomuuid = virtinst.Guest.generate_uuid(testconn)
    assert randomuuid != testuuid
    assert len(randomuuid) == len(testuuid)

    testmac = virtinst.DeviceInterface.generate_mac(predconn)
    randommac = virtinst.DeviceInterface.generate_mac(testconn)
    qemumac = virtinst.DeviceInterface.generate_mac(kvmconn)
    assert randommac != testmac
    assert qemumac != testmac
    assert len(randommac) == len(testmac)

    # Ensure check_mac_in_use doesn't error on None
    virtinst.DeviceInterface.check_mac_in_use(predconn, None)


def test_misc_support_cornercases():
    """
    Test support.py corner cases
    """
    conn = utils.URIs.open_testdefault_cached()
    try:
        conn.lookupByName("foobar-idontexist")
    except Exception as e:
        if not conn.support.is_libvirt_error_no_domain(e):
            raise


def test_misc_osxml_cornercases():
    """
    Test OSXML corner cases
    """
    conn = utils.URIs.open_testdefault_cached()
    guest = virtinst.Guest(conn)
    guest.os.set_initargs_string("foo bar")
    guest.os.set_initargs_string("baz wibble")
    assert [i.val for i in guest.os.initargs] == ["baz", "wibble"]


def test_misc_cpu_cornercases():
    """
    Hit the validation paths for default HOST_MODEL_ONLY
    """
    kvmconn = utils.URIs.open_kvm()
    guest = virtinst.Guest(kvmconn)
    guest.x86_cpu_default = guest.cpu.SPECIAL_MODE_HOST_MODEL_ONLY
    guest.set_defaults(guest)
    assert guest.cpu.model == "Skylake-Client-noTSX-IBRS"

    # pylint: disable=protected-access
    guest.cpu.model = "idontexist"
    guest.cpu._validate_default_host_model_only(guest)
    assert guest.cpu.model is None


def test_misc_meter():
    """
    Test coverage of our urlgrabber meter copy
    """
    # pylint: disable=protected-access
    from virtinst import _progresspriv

    def _test_meter_values(m, startval=10000, text="Meter text test"):
        with unittest.mock.patch("time.time", return_value=1.0):
            m.start(text, startval)
        with unittest.mock.patch("time.time", return_value=1.1):
            m.update(0)
        with unittest.mock.patch("time.time", return_value=1.5):
            m.update(0)
        with unittest.mock.patch("time.time", return_value=2.0):
            m.update(100)
        with unittest.mock.patch("time.time", return_value=3.0):
            m.update(200)
        with unittest.mock.patch("time.time", return_value=4.0):
            m.update(2000)
        with unittest.mock.patch("time.time", return_value=5.0):
            m.update(4000)
        with unittest.mock.patch("time.time", return_value=5.1):
            m.update(4500)
        with unittest.mock.patch("time.time", return_value=5.5):
            m.end()

    # Basic output testing
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _test_meter_values(meter)
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter1.txt"))

    # Fake having a longer terminal, it affects output a bit
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _progresspriv._term_width_val = 120
    _test_meter_values(meter)
    _progresspriv._term_width_val = 80
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter2.txt"))

    # meter with size=None
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _test_meter_values(meter, None)
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter3.txt"))

    # meter with size=None and small terminal size
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _progresspriv._term_width_val = 11
    _test_meter_values(meter, None, "1234567890")
    assert meter.re.fraction_read() is None
    _progresspriv._term_width_val = 80
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter4.txt"))

    # meter with size exceeded by the update() values
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _test_meter_values(meter, 200)
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter5.txt"))

    # meter with size 0
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _test_meter_values(meter, 0)
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter6.txt"))

    def _test_meter_zero(m, startval=0, text="Meter text test"):
        with unittest.mock.patch("time.time", return_value=1.0):
            m.start(text, startval)
        with unittest.mock.patch("time.time", return_value=3.0):
            m.update(0)
        with unittest.mock.patch("time.time", return_value=3.1):
            m.end()

    # meter with size 0 and startval size 0
    meter = _progresspriv.TextMeter(output=io.StringIO())
    _test_meter_zero(meter, 0)
    out = meter.output.getvalue().replace("\r", "\n")
    utils.diff_compare(out, os.path.join(utils.DATADIR, "meter", "meter-zero.txt"))

    # BaseMeter coverage
    meter = _progresspriv.BaseMeter()
    _test_meter_values(meter)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

import tests.utils

import virtinst


def _skipIfTestsFailed():
    if tests.utils.TESTCONFIG.skip_checkprops:
        pytest.skip("Other tests failed or were skipped, don't do prop checks")


def testCheckXMLBuilderProps():
    """
    If a certain environment variable is set, XMLBuilder tracks
    every property registered and every one of those that is
    actually altered. The test suite sets that env variable.
    If no tests failed or were skipped, we check to ensure the
    test suite is tickling every XML property
    """
    _skipIfTestsFailed()

    # pylint: disable=protected-access
    fail = [p for p in virtinst.xmlbuilder._allprops if p not in virtinst.xmlbuilder._seenprops]
    if fail:
        msg = "\n\n"
        msg += "\n".join(str(a) for a in fail) + "\n\n"
        msg += (
            "This means that there are XML properties that are\n"
            "untested in the test suite. This could be caused\n"
            "by a previous test suite failure, or if you added\n"
            "a new property and didn't extend the test suite.\n"
            "Look into extending test_cli.py and/or test_xmlparse.py."
        )
        pytest.fail(msg)


def testCheckCLISuboptions():
    """
    Track which command line suboptions and aliases we actually hit with
    the test suite.
    """
    _skipIfTestsFailed()

    # pylint: disable=protected-access
    from virtinst import cli

    unchecked = cli._SuboptChecker.get_unseen()
    if unchecked:
        msg = "\n\n"
        msg += "\n".join(sorted(a for a in unchecked)) + "\n\n"
        msg += (
            "These command line arguments or aliases are not checked\n"
            "in the test suite. Please test them.\n"
            "Total unchecked arguments: %s" % len(unchecked)
        )
        pytest.fail(msg)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import tests.utils
from . import lib


class _CloneRow:
    """
    Helper class for interacting with the clone row
    """

    def __init__(self, *args):
        self.chkcell = args[2]
        self.txtcell = args[5]

        self.is_cloneable = self.chkcell.showing
        self.is_share_requested = not self.is_cloneable or not self.chkcell.checked
        self.is_clone_requested = not self.is_share_requested

    def check_in_text(self, substr):
        lib.utils.check(lambda: substr in self.txtcell.text)

    def select(self):
        self.txtcell.click()


def _get_all_rows(win):
    slist = win.find("storage-list")

    def pred(node):
        return node.roleName == "table cell"

    cells = slist.findChildren(pred, isLambda=True)

    idx = 0
    rows = []
    cellcount = 6
    while idx < len(cells):
        rows.append(_CloneRow(*cells[idx : idx + cellcount]))
        idx += cellcount
        # Skip the next row which is always a separator
        idx += cellcount
    return rows


##############################################
# UI tests for virt-manager's CloneVM wizard #
##############################################


def testCloneSimple(app):
    # Disable predictable so UUID generation doesn't collide
    uri = tests.utils.URIs.test_full.replace(",predictable", "")
    app.uri = uri

    # Clone 'test-clone-simple' which is the most basic case
    # Cancel, and reopen
    win = app.manager_open_clone("test-clone-simple")
    win.find("Cancel", "push button").click()
    lib.utils.check(lambda: not win.showing)

    # Do default clone
    win = app.manager_open_clone("test-clone-simple")
    rows = _get_all_rows(win)
    assert len(rows) == 1
    assert rows[0].is_clone_requested
    rows[0].check_in_text("test-clone-simple.img")

    win.find("Clone", "push button").click()
    lib.utils.check(lambda: not win.showing)

    # Check path was generated correctly
    win = app.manager_open_clone("test-clone-simple-clone")
    rows = _get_all_rows(win)
    assert len(rows) == 1
    assert rows[0].is_clone_requested
    rows[0].check_in_text("test-clone-simple-clone.img")

    # Share storage and deal with warnings
    rows[0].chkcell.click()
    rows[0].check_in_text("Share disk with")
    # Do 'cancel' first
    win.find("Clone", "push button").click()
    app.click_alert_button("cause data to be overwritten", "Cancel")
    lib.utils.check(lambda: win.active)
    win.find("Clone", "push button").click()
    app.click_alert_button("cause data to be overwritten", "OK")
    lib.utils.check(lambda: not win.active)

    # Verify the new VM shared storage
    win = app.manager_open_clone("test-clone-simple-clone1")
    rows = _get_all_rows(win)
    assert len(rows) == 1
    rows[0].check_in_text("test-clone-simple-clone.img")


def testCloneMulti(app):
    # Clone 'test-clone', check some results, make sure clone works
    manager = app.topwin
    manager.window_maximize()

    # Shutdown this VM to prep for later
    manager.find("test-many-devices").click()
    sbutton = manager.find("Shut Down", "push button")
    sbutton.click()
    lib.utils.check(lambda: not sbutton.sensitive)

    # Do a basic clone
    win = app.manager_open_clone("test-clone")
    win.find("Clone", "push button").click()
    lib.utils.check(lambda: not win.showing)
    manager.find("test-clone1", "table cell")

    # Check test-many-devices which will not work, but confirm
    # it errors gracefully
    win = app.manager_open_clone("test-many-devices")
    win.find("Clone", "push button").click()
    app.click_alert_button("relative.sock", "Cancel")

    # Ensure disconnecting will close the dialog
    manager.grab_focus()
    app.manager_conn_disconnect("test testdriver.xml")
    lib.utils.check(lambda: not win.showing)


def testCloneStorageChange(app):
    # Disable predictable so UUID generation doesn't collide
    uri = tests.utils.URIs.test_full.replace(",predictable", "")
    app.uri = uri

    # Trigger some error handling scenarios
    win = app.manager_open_clone("test-clone-simple")
    newname = "test-aaabbb"
    win.find("Name:", "text").set_text(newname)
    win.find("Clone", "push button").click()
    lib.utils.check(lambda: not win.showing)

    win = app.manager_open_clone(newname)
    row = _get_all_rows(win)[0]
    row.check_in_text(newname)
    oldnewname = newname
    newname = "test-aaazzzzbbb"
    win.find("Name:", "text").set_text(newname)
    row.select()

    win.find("Details", "push button").click()
    stgwin = app.root.find("Change storage path", "dialog")
    pathtxt = stgwin.find(None, "text", "New Path:")
    lib.utils.check(lambda: newname in pathtxt.text)
    stgwin.find("Browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "iso-vol")
    lib.utils.check(lambda: "iso-vol" in pathtxt.text)
    stgwin.find("OK").click()
    app.click_alert_button("overwrite the existing", "No")
    lib.utils.check(lambda: stgwin.showing)
    stgwin.find("OK").click()
    app.click_alert_button("overwrite the existing", "Yes")
    lib.utils.check(lambda: not stgwin.showing)
    # Can't clone onto existing storage volume
    win.find("Clone", "push button").click()
    app.click_alert_button(".*Clone onto existing.*", "Close")

    # Reopen dialog and request to share it
    win.find("Details", "push button").click()
    stgwin = app.root.find("Change storage path", "dialog")
    chkbox = stgwin.find("Create a new", "check")
    lib.utils.check(lambda: chkbox.checked)
    chkbox.click()

    # Cancel and reopen, confirm changes didn't stick
    stgwin.find("Cancel").click()
    lib.utils.check(lambda: not stgwin.showing)
    win.find("Details", "push button").click()
    stgwin = app.root.find("Change storage path", "dialog")
    chkbox = stgwin.find("Create a new", "check")
    lib.utils.check(lambda: chkbox.checked)
    # Requesting sharing again and exit
    chkbox.click()
    stgwin.find("OK").click()
    lib.utils.check(lambda: not stgwin.active)

    # Finish install, verify storage was shared
    win.find("Clone", "push button").click()
    app.click_alert_button("cause data to be overwritten", "OK")
    lib.utils.check(lambda: not win.active)
    win = app.manager_open_clone(newname)
    row = _get_all_rows(win)[0].check_in_text(oldnewname)


def testCloneError(app):
    # Trigger some error handling scenarios
    win = app.manager_open_clone("test-clone-full")
    win.find("Clone", "push button").click()
    app.click_alert_button("not enough free space", "Close")
    win.window_close()

    win = app.manager_open_clone("test-clone-simple")
    badname = "test/foo"
    win.find("Name:", "text").set_text(badname)
    rows = _get_all_rows(win)
    rows[0].chkcell.click()
    rows[0].check_in_text("Share disk with")
    win.find("Clone", "push button").click()
    win.find("Clone", "push button").click()
    app.click_alert_button("cause data to be overwritten", "OK")
    app.click_alert_button(badname, "Close")
    lib.utils.check(lambda: win.active)


def testCloneNonmanaged(app):
    # Verify unmanaged clone actual works
    import tempfile

    tmpsrc = tempfile.NamedTemporaryFile()
    tmpdst = tempfile.NamedTemporaryFile()

    open(tmpsrc.name, "w").write(__file__)

    app.open(xmleditor_enabled=True)
    manager = app.topwin

    win = app.manager_open_details("test-clone-simple")
    win.find("IDE Disk 1", "table cell").click()
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    origpath = "/pool-dir/test-clone-simple.img"
    newpath = tmpsrc.name
    xmleditor.set_text(xmleditor.text.replace(origpath, newpath))
    win.find("config-apply").click()
    win.find("Details", "page tab").click()
    disksrc = win.find("disk-source-path")
    lib.utils.check(lambda: disksrc.text == newpath)
    win.window_close()

    lib.utils.check(lambda: manager.active)
    win = app.manager_open_clone("test-clone-simple")
    row = _get_all_rows(win)[0]
    row.check_in_text(tmpsrc.name)
    row.select()

    win.find("Details", "push button").click()
    stgwin = app.root.find("Change storage path", "dialog")
    pathtxt = stgwin.find(None, "text", "New Path:")
    os.unlink(tmpdst.name)
    pathtxt.set_text(tmpdst.name)
    stgwin.find("OK").click()
    win.find("Clone", "push button").click()
    lib.utils.check(lambda: not win.active)
    lib.utils.check(lambda: os.path.exists(tmpdst.name))

    assert open(tmpsrc.name).read() == open(tmpdst.name).read()



# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import sys


_alldistros = {}

DEVFEDORA_URL = "http://dl.fedoraproject.org/pub/fedora/linux/development/%s/Server/%s/os/"
FEDORA_URL = "http://dl.fedoraproject.org/pub/fedora/linux/releases/%s/Server/%s/os/"

(WARN_RHEL5, WARN_DEBIAN, WARN_FEDORA) = range(1, 4)


def prompt():
    sys.stdout.write("(press enter to continue)")
    sys.stdout.flush()
    return sys.stdin.readline()


KSOLD = "tests/data/inject/old-kickstart.ks"
KSNEW = "tests/data/inject/new-kickstart.ks"
PRESEED = "tests/data/inject/preseed.cfg"


class Distro:
    def __init__(self, name, url, filename, warntype=WARN_FEDORA):
        self.name = name
        self.url = url
        self.warntype = warntype
        self.filename = filename

        self.kernel = None
        self.initrd = None


def _add(*args, **kwargs):
    _d = Distro(*args, **kwargs)
    _alldistros[_d.name] = _d


_add("centos5.11", "http://vault.centos.org/5.11/os/x86_64/", warntype=WARN_RHEL5, filename=KSOLD)
_add("centos6.10", "http://vault.centos.org/6.10/os/x86_64", warntype=WARN_RHEL5, filename=KSOLD)
_add("centos7latest", "http://ftp.linux.ncsu.edu/pub/CentOS/7/os/x86_64/", filename=KSNEW)
_add(
    "centos8stream",
    "http://ftp.linux.ncsu.edu/pub/CentOS/8-stream/BaseOS/x86_64/os/",
    filename=KSNEW,
)
_add("fedora35", FEDORA_URL % ("29", "x86_64"), filename=KSNEW)
_add("fedora36", DEVFEDORA_URL % ("35", "x86_64"), filename=KSNEW)
_add(
    "debian9",
    "http://ftp.us.debian.org/debian/dists/stretch/main/installer-amd64/",
    filename=PRESEED,
    warntype=WARN_DEBIAN,
)
_add(
    "debian11",
    "http://ftp.us.debian.org/debian/dists/bullseye/main/installer-amd64/",
    filename=PRESEED,
    warntype=WARN_DEBIAN,
)


def _test_distro(distro):
    os.system("clear")
    print("\n")
    if distro.warntype == WARN_RHEL5:
        print("RHEL5, RHEL6, Fedora < 17: You'll get an error about a ")
        print("bogus bootproto ITREADTHEKICKSTART. This means anaconda ")
        print("read our busted kickstart.")
    elif distro.warntype == WARN_DEBIAN:
        print(
            "Debian: Won't ask any questions, will autoconfig network, "
            "then print a big red text box about a bad mirror config."
        )
    elif distro.warntype == WARN_FEDORA:
        print("RHEL, Fedora >= 17: Chokes on the bogus URI in the early ")
        print("console screen when fetching the installer squashfs image.")

    os.environ.pop("VIRTINST_TEST_SUITE", None)
    os.environ["VIRTINST_INITRD_TEST"] = "1"

    if distro.warntype == WARN_DEBIAN:
        append = "auto=true"
    else:
        append = '"ks=file:/%s"' % os.path.basename(distro.filename)
    cmd = (
        "./virt-install --connect qemu:///system "
        "--name __virtinst__test__initrd__ --ram 2048 "
        "--transient --destroy-on-exit --disk none "
        "--location %s --initrd-inject %s "
        "--install kernel_args=%s,kernel_args_overwrite=yes" % (distro.url, distro.filename, append)
    )
    print("\n\n" + cmd)
    os.system(cmd)


def _print_intro():
    print(
        """


This is an interactive test suite.

We are going to launch various transient virt-installs, using initrd
injections, that will cause installs to quickly fail. Look for the
failure pattern to confirm that initrd injections are working as expected.

"""
    )
    prompt()


def _build_testfunc(dobj, do_setup):
    def testfunc():
        if do_setup:
            _print_intro()
        _test_distro(dobj)

    return testfunc


def _make_tests():
    idx = 0
    for dname, dobj in _alldistros.items():
        idx += 1
        name = "testInitrd%.3d_%s" % (idx, dname)

        do_setup = idx == 1
        testfunc = _build_testfunc(dobj, do_setup)
        globals()[name] = testfunc


_make_tests()



# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

import virtinst

from tests import utils


DATADIR = utils.DATADIR + "/xmlparse/"


####################
# Helper functions #
####################


def _sanitize_file_xml(xml):
    # s/"/'/g from generated XML, matches what libxml dumps out
    # This won't work all the time, but should be good enough for testing
    return xml.replace("'", '"')


def _alter_compare(conn, actualXML, outfile):
    utils.diff_compare(actualXML, outfile)
    utils.check_create(conn, actualXML)


def _set_and_check(obj, param, initval, *args):
    """
    Check expected initial value obj.param == initval, then
    set newval, and make sure it is returned properly
    """
    curval = virtinst.xmlutil.get_prop_path(obj, param)
    assert initval == curval

    for newval in args:
        virtinst.xmlutil.set_prop_path(obj, param, newval)
        curval = virtinst.xmlutil.get_prop_path(obj, param)
        assert newval == curval


def _make_checker(obj):
    def check(name, initval, *args):
        return _set_and_check(obj, name, initval, *args)

    return check


def _gen_outfile_path(basename):
    """
    Returns relative path to the file containing the expected XML
    output

    """
    return DATADIR + "{!s}-out.xml".format(basename)


def _get_test_content(conn, basename):
    infile = DATADIR + "%s-in.xml" % basename
    outfile = _gen_outfile_path(basename)
    guest = virtinst.Guest(conn, parsexml=open(infile).read())
    return guest, outfile


##############
# Test cases #
##############


def testAlterGuest():
    """
    Test changing Guest() parameters after parsing
    """
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "change-guest")

    check = _make_checker(guest)

    # Check specific vcpu_current behavior
    check("vcpus", 5, 10)
    assert guest.vcpu_current is None
    check("vcpu_current", None, 15)
    guest.vcpus = 12
    assert guest.vcpu_current == 12
    guest.vcpu_current = 10

    check("name", "TestGuest", "change_name")
    check("id", None, 1234)
    check("description", None, "Hey desc changed&")
    check("title", None, "Hey title changed!")
    check("vcpu_cpuset", "1-3", "1-8,^6", "1-5,15")
    check("memory", 409600, 512000)
    check("currentMemory", 204800, 1024000)
    check("memory", 1024000, 2048000)
    check("uuid", "12345678-1234-1234-1234-123456789012", "11111111-2222-3333-4444-555555555555")
    check("emulator", "/usr/lib/xen/bin/qemu-dm", "/usr/binnnn/fooemu")
    check("type", "kvm", "test")
    check("bootloader", None, "pygrub")
    check("on_poweroff", "destroy", "restart")
    check("on_reboot", "restart", "destroy")
    check("on_crash", "destroy", "restart")
    check("on_lockfailure", "poweroff", "restart")

    check = _make_checker(guest._metadata.libosinfo)  # pylint: disable=protected-access
    check("os_id", "http://fedoraproject.org/fedora/17")
    guest.set_os_name("fedora10")
    check("os_id", "http://fedoraproject.org/fedora/10")
    assert guest.osinfo.name == "fedora10"
    guest.set_os_name("generic")
    check("os_id", None, "frib")
    assert guest.osinfo.name == "generic"

    check = _make_checker(guest.clock)
    check("offset", "utc", "localtime")
    guest.clock.remove_child(guest.clock.timers[0])
    check = _make_checker(guest.clock.timers[0])
    check("name", "pit", "rtc")
    check("tickpolicy", "delay", "merge")
    timer = guest.clock.timers.add_new()
    check = _make_checker(timer)
    check("name", None, "hpet")
    check("present", None, False)

    check = _make_checker(guest.pm)
    check("suspend_to_mem", False, True)
    check("suspend_to_disk", None, False)

    check = _make_checker(guest.os)
    check("os_type", "hvm", "xen")
    check("arch", "i686", None)
    check("machine", "foobar", "pc-0.11")
    check("loader", None, "/foo/loader")
    check("init", None, "/sbin/init")
    check("bootorder", ["hd"], ["fd"])
    check("bootmenu_enable", None, False)
    check("bootmenu_timeout", None, 30000)
    check("bios_useserial", None, True)
    check("bios_rebootTimeout", None, -1)
    check("kernel", None)
    check("initrd", None)
    check("kernel_args", None)

    guest.os.set_initargs_string("foo 'bar baz' frib")
    assert [i.val for i in guest.os.initargs] == ["foo", "bar baz", "frib"]

    check = _make_checker(guest.features)
    check("acpi", True, False)
    check("apic", True, True)
    check("eoi", None, True)
    check("pae", False, False)
    check("viridian", False, True)
    check("hap", False, False)
    check("privnet", False, False)
    check("hyperv_relaxed", None, True)
    check("hyperv_vapic", False, None)
    check("hyperv_spinlocks", True, True)
    check("hyperv_spinlocks_retries", 12287, 54321)
    check("vmport", False, True)
    check("vmcoreinfo", None, True)
    check("kvm_hidden", None, True)
    check("pvspinlock", None, True)
    check("gic_version", None, False)
    check("ioapic_driver", None, "qemu")
    check("kvm_pv_ipi", None, False)

    check = _make_checker(guest.cpu)
    check("match", "exact", "strict")
    guest.cpu.set_model(guest, "qemu64")
    check("model", "qemu64")
    check("vendor", "Intel", "qemuvendor")
    check("topology.threads", 2, 1)
    check("topology.cores", 5, 3)
    guest.cpu.topology.sockets = 4.0
    check("topology.sockets", 4)

    check = _make_checker(guest.cpu.features[0])
    check("name", "x2apic")
    check("policy", "force", "disable")
    rmfeat = guest.cpu.features[3]
    guest.cpu.remove_child(rmfeat)
    assert rmfeat.get_xml() == """<feature name="foo" policy="bar"/>\n"""
    guest.cpu.add_feature("addfeature")

    check = _make_checker(guest.numatune)
    check("memory_mode", "interleave", "strict", None)
    check("memory_nodeset", "1-5,^3,7", "2,4,6")

    check = _make_checker(guest.memtune)
    check("hard_limit", None, 1024, 2048)
    check("soft_limit", None, 100, 200)
    check("swap_hard_limit", None, 300, 400)
    check("min_guarantee", None, 400, 500)

    check = _make_checker(guest.blkiotune)
    check("weight", None, 100, 200)
    check = _make_checker(guest.blkiotune.devices.add_new())
    check("weight", None, 300)
    check("path", None, "/home/1.img")

    check = _make_checker(guest.idmap)
    check("uid_start", None, 0)
    check("uid_target", None, 1000)
    check("uid_count", None, 10)
    check("gid_start", None, 0)
    check("gid_target", None, 1000)
    check("gid_count", None, 10)

    check = _make_checker(guest.resource)
    check("partition", None, "/virtualmachines/production")

    check = _make_checker(guest.devices.memballoon[0])
    check("model", "virtio", "none")

    check = _make_checker(guest.memoryBacking)
    check("hugepages", False, True)
    check("nosharepages", False, True)
    check("locked", False, True)

    page = guest.memoryBacking.pages.add_new()
    check = _make_checker(page)
    check("size", None, 1)
    check("unit", None, "G")

    assert guest.is_full_os_container() is False
    _alter_compare(conn, guest.get_xml(), outfile)


def testAlterCpuMode():
    conn = utils.URIs.open_testdefault_cached()
    xml = open(DATADIR + "change-cpumode-in.xml").read()
    outfile = DATADIR + "change-cpumode-out.xml"
    conn = utils.URIs.openconn(utils.URIs.kvm_x86)
    guest = virtinst.Guest(conn, xml)
    check = _make_checker(guest.cpu)

    guest.cpu.model = "foo"
    check("mode", "host-passthrough")
    guest.cpu.check_security_features(guest)
    check("secure", False)
    guest.cpu.set_special_mode(guest, "host-model")
    check("mode", "host-model")
    guest.cpu.check_security_features(guest)
    check("secure", False)
    guest.cpu.set_model(guest, "qemu64")
    check("model", "qemu64")
    guest.cpu.check_security_features(guest)
    check("secure", False)

    # Test actually filling in security values, and removing them
    guest.cpu.secure = True
    guest.cpu.set_model(guest, "Skylake-Client-IBRS")
    guest.cpu.check_security_features(guest)
    check("secure", True)
    guest.cpu.set_model(guest, "EPYC-IBPB")
    guest.cpu.check_security_features(guest)
    check("secure", True)
    guest.cpu.secure = False
    guest.cpu.set_model(guest, "Skylake-Client-IBRS")
    guest.cpu.check_security_features(guest)
    check("secure", False)
    _alter_compare(conn, guest.get_xml(), outfile)

    # Hits a codepath when domcaps don't provide the needed info
    emptyconn = utils.URIs.open_testdefault_cached()
    guest = virtinst.Guest(emptyconn, xml)
    guest.cpu.check_security_features(guest)
    assert guest.cpu.secure is False


def testAlterDisk():
    """
    Test changing DeviceDisk() parameters after parsing
    """
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "change-disk")

    def _get_disk(target):
        for disk in guest.devices.disk:
            if disk.target == target:
                return disk

    disk = _get_disk("hda")
    check = _make_checker(disk)
    assert disk.get_source_path() == "/tmp/test.img"
    disk.set_source_path("/dev/foo/null")
    disk.sync_path_props()
    check("driver_name", None, "test")
    check("driver_type", None, "raw")
    check("serial", "WD-WMAP9A966149", "frob")
    check("wwn", None, "123456789abcdefa")
    check("bus", "ide", "usb")
    check("removable", None, False, True)

    disk = guest.devices.disk[1]
    check = _make_checker(disk.seclabels[1])
    check("model", "dac")
    check("relabel", None, True)
    check("label", None, "foo-my-label")

    disk = _get_disk("hdc")
    check = _make_checker(disk)
    check("type", "block", "dir", "file", "block")
    assert disk.get_source_path() == "/dev/null"
    disk.set_source_path(None)
    disk.sync_path_props()
    check("device", "cdrom", "floppy")
    check("read_only", True, False)
    check("target", "hdc", "fde")
    check("bus", "ide", "fdc")
    check("error_policy", "stop", None)

    disk = _get_disk("hdd")
    check = _make_checker(disk)
    check("type", "block")
    check("device", "lun")

    disk = _get_disk("sda")
    check = _make_checker(disk)
    disk.set_source_path("http://[1:2:3:4:5:6:7:8]:1122/my/file")
    disk.sync_path_props()

    disk = _get_disk("fda")
    check = _make_checker(disk)
    disk.set_source_path("/pool-dir/default-vol")
    disk.sync_path_props()
    check("startup_policy", None, "optional")
    check("shareable", False, True)
    check("driver_cache", None, "writeback")
    check("driver_io", None, "threads")
    check("driver_io", "threads", "native")
    check("driver_discard", None, "unmap")
    check("driver_detect_zeroes", None, "unmap")
    check("iotune_ris", 1, 0)
    check("iotune_rbs", 2, 0)
    check("iotune_wis", 3, 0)
    check("iotune_wbs", 4, 0)
    check("iotune_tis", None, 5)
    check("iotune_tbs", None, 6)
    check = _make_checker(disk.boot)
    check("order", None, 7, None)

    disk = _get_disk("vdb")
    check = _make_checker(disk)
    check("source.pool", "defaultPool", "anotherPool")
    check("source.volume", "foobar", "newvol")

    disk = _get_disk("vdc")
    check = _make_checker(disk)
    check("source.protocol", "rbd", "gluster")
    check("source.name", "pool/image", "new-val/vol")
    hcheck = _make_checker(disk.source.hosts[0])
    hcheck("name", "mon1.example.org", "diff.example.org")
    hcheck("port", 6321, 1234)
    assert disk.get_source_path() == "gluster://diff.example.org:1234/new-val/vol"

    disk = _get_disk("vdd")
    check = _make_checker(disk)
    hcheck = _make_checker(disk.source.hosts[0])
    check("source.protocol", "nbd")
    hcheck("transport", "unix")
    hcheck("socket", "/var/run/nbdsock")
    assert disk.get_source_path() == "nbd+unix:///var/run/nbdsock"

    _alter_compare(conn, guest.get_xml(), outfile)


def testAlterDevicesBootorder():
    conn = utils.URIs.open_testdefault_cached()
    basename = "change-devices-bootorder"
    guest, outfile = _get_test_content(conn, basename)
    disk_1 = guest.devices.disk[0]
    disk_2 = guest.devices.disk[1]
    disk_3 = guest.devices.disk[2]
    disk_4 = guest.devices.disk[3]
    iface_1 = guest.devices.interface[0]
    iface_2 = guest.devices.interface[1]
    redirdev_1 = guest.devices.redirdev[0]

    assert guest.os.bootorder == ["hd"]
    assert disk_1.boot.order is None
    assert disk_2.boot.order == 10
    assert disk_3.boot.order == 10
    assert disk_4.boot.order == 1
    assert iface_1.boot.order == 2
    assert iface_2.boot.order is None
    assert redirdev_1.boot.order == 3

    guest.reorder_boot_order(disk_1, 1)

    assert guest.os.bootorder == []
    assert disk_1.boot.order == 1
    assert disk_2.boot.order == 10
    assert disk_3.boot.order == 10
    # verify that the used algorithm preserves the order of
    # records with equal boot indices
    assert disk_2 is guest.devices.disk[1]
    assert disk_3 is guest.devices.disk[2]
    assert disk_4.boot.order == 2
    assert iface_1.boot.order == 3
    assert iface_2.boot.order is None
    assert redirdev_1.boot.order == 4

    with pytest.raises(RuntimeError, match="unsupported configuration"):
        _alter_compare(conn, guest.get_xml(), outfile)

    guest.reorder_boot_order(disk_2, 10)
    assert disk_2.boot.order == 10
    assert disk_3.boot.order == 11
    assert disk_2 is guest.devices.disk[1]
    assert disk_3 is guest.devices.disk[2]

    outfile = _gen_outfile_path("change-devices-bootorder-fixed")
    _alter_compare(conn, guest.get_xml(), outfile)


def testSingleDisk():
    conn = utils.URIs.open_testdefault_cached()
    xml = (
        """<disk type="file" device="disk"><source file="/a.img"/>\n"""
        """<target dev="hda" bus="ide"/></disk>\n"""
    )
    conn = utils.URIs.open_testdefault_cached()
    d = virtinst.DeviceDisk(conn, parsexml=xml)
    _set_and_check(d, "target", "hda", "hdb")
    assert xml.replace("hda", "hdb") == d.get_xml()


def testAlterChars():
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "change-chars")

    serial1 = guest.devices.serial[0]
    serial2 = guest.devices.serial[1]
    parallel1 = guest.devices.parallel[0]
    parallel2 = guest.devices.parallel[1]
    console1 = guest.devices.console[0]
    console2 = guest.devices.console[1]
    channel1 = guest.devices.channel[0]
    channel2 = guest.devices.channel[1]
    channel3 = guest.devices.channel[2]

    check = _make_checker(serial1)
    check("type", "null", "udp")
    check("source.bind_host", None, "example.com")
    check("source.bind_service", None, 66)
    check("source.connect_host", None, "example.com.uk")
    check("source.connect_service", None, 77)

    check = _make_checker(serial2)
    check("type", "tcp")
    check("source.protocol", "telnet", "raw")
    check("source.mode", "bind", "connect")

    check = _make_checker(parallel1)
    check("source.mode", "bind")
    check("source.path", "/tmp/foobar", None)
    check("type", "unix", "pty")

    check = _make_checker(parallel2)
    check("type", "udp")
    check("source.bind_service", 1111, 1357)
    check("source.bind_host", "my.bind.host", "my.foo.host")
    check("source.connect_service", 2222, 7777)
    check("source.connect_host", "my.source.host", "source.foo.host")

    check = _make_checker(console1)
    check("type", "pty")
    check("target_type", None)

    check = _make_checker(console2)
    check("type", "file")
    check("source.path", "/tmp/foo.img", None)
    check("source.path", None, "/root/foo")
    check("target_type", "virtio")
    check("target_state", None, "connected")

    check = _make_checker(channel1)
    check("type", "pty")
    check("target_type", "virtio", "bar", "virtio")
    check("target_name", "foo.bar.frob", "test.changed")

    check = _make_checker(channel2)
    check("type", "unix", "foo", "unix")
    check("target_type", "guestfwd")
    check("target_address", "1.2.3.4", "5.6.7.8")
    check("target_port", 4567, 1199)

    check = _make_checker(channel3)
    check("type", "spiceport")
    check("source.channel", "org.spice-space.webdav.0", "test.1")
    check("target_type", "virtio")
    check("target_name", "org.spice-space.webdav.0", "test.2")
    assert channel3.get_xml_id() == "./devices/channel[3]"
    assert channel3.get_xml_idx() == 2

    _alter_compare(conn, guest.get_xml(), outfile)


def testAlterNics():
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "change-nics")

    dev1 = guest.devices.interface[0]
    dev2 = guest.devices.interface[1]
    dev3 = guest.devices.interface[2]
    dev4 = guest.devices.interface[3]
    dev5 = guest.devices.interface[4]

    check = _make_checker(dev1)
    check("type", "user")
    check("model", None, "vmxnet3")
    check(
        "source",
        None,
        None,
    )
    check("macaddr", "22:11:11:11:11:11", "AA:AA:AA:AA:AA:AA")
    check("filterref", None, "foo")

    check = _make_checker(dev2)
    check("source", "default", None)
    check("type", "network", "bridge")
    check("source", None, "newbr0")
    check("model", "e1000", "virtio")

    check = _make_checker(dev3)
    check("type", "bridge")
    check("source", "foobr0", "newfoo0")
    check("macaddr", "22:22:22:22:22:22")
    check("target_dev", None, "test1")

    check = _make_checker(dev4)
    check("type", "ethernet")
    check("target_dev", "nic02", "nic03")
    check("target_dev", "nic03", None)

    check = _make_checker(dev5)
    check("type", "direct")
    check("source", "eth0.1")
    check("source_mode", "vepa", "bridge")
    check("portgroup", None, "sales")
    check("driver_name", None, "vhost")
    check("driver_queues", None, 5)

    virtualport = dev5.virtualport
    check = _make_checker(virtualport)
    check("type", "802.1Qbg", "foo", "802.1Qbg")
    check("managerid", 12, 11)
    check("typeid", 1193046, 1193047)
    check("typeidversion", 1, 2)
    check(
        "instanceid", "09b11c53-8b5c-4eeb-8f00-d84eaa0aaa3b", "09b11c53-8b5c-4eeb-8f00-d84eaa0aaa4f"
    )

    _alter_compare(conn, guest.get_xml(), outfile)


def testQEMUXMLNS():
    kvmconn = utils.URIs.open_kvm()
    basename = "change-xmlns-qemu"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    guest = virtinst.Guest(kvmconn, parsexml=open(infile).read())

    check = _make_checker(guest.xmlns_qemu.args[0])
    check("value", "-somearg", "-somenewarg")
    check = _make_checker(guest.xmlns_qemu.args[1])
    check("value", "bar,baz=wib wob", "testval")
    guest.xmlns_qemu.args.add_new().value = "additional-arg"
    arg0 = guest.xmlns_qemu.args[0]
    guest.xmlns_qemu.remove_child(guest.xmlns_qemu.args[0])
    x = '<qemu:arg xmlns:qemu="http://libvirt.org/schemas/domain/qemu/1.0" value="-somenewarg"/>\n'
    assert arg0.get_xml() == x

    check = _make_checker(guest.xmlns_qemu.envs[0])
    check("name", "SOMEENV")
    check("value", "foo=bar baz,foo")
    env = guest.xmlns_qemu.envs.add_new()
    env.name = "DISPLAY"
    env.value = "1:2"

    _alter_compare(kvmconn, guest.get_xml(), outfile)


def testAddRemoveDevices():
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "add-devices")

    # Basic removal of existing device
    rmdev = guest.devices.disk[2]
    guest.remove_device(rmdev)

    # Basic device add
    d = virtinst.DeviceWatchdog(conn)
    d.set_defaults(guest)
    guest.add_device(d)

    # Test adding device with child properties (address value)
    adddev = virtinst.DeviceInterface(conn)
    adddev.type = "network"
    adddev.source = "default"
    adddev.macaddr = "1A:2A:3A:4A:5A:6A"
    adddev.address.type = "spapr-vio"
    adddev.set_defaults(guest)

    # Test adding and removing the same device
    guest.add_device(adddev)
    guest.remove_device(adddev)
    guest.add_device(adddev)

    # Test adding device built from parsed XML
    guest.add_device(virtinst.DeviceSound(conn, parsexml="""<sound model='pcspk'/>"""))

    _alter_compare(conn, guest.get_xml(), outfile)


def testChangeKVMMedia():
    kvmconn = utils.URIs.open_kvm()
    guest, outfile = _get_test_content(kvmconn, "change-media")

    disk = guest.devices.disk[0]
    disk.set_source_path("/pool-dir/default-vol")
    disk.sync_path_props()

    disk = guest.devices.disk[1]
    disk.set_source_path("/pool-dir/default-vol")
    assert disk.get_source_path() == "/pool-dir/default-vol"
    disk.set_source_path("/dev/pool-logical/diskvol1")
    disk.sync_path_props()

    disk = guest.devices.disk[2]
    disk.set_source_path("/dev/pool-logical/diskvol1")
    disk.sync_path_props()

    disk = guest.devices.disk[3]
    disk.set_source_path("/pool-dir/default-vol")
    disk.sync_path_props()

    disk = guest.devices.disk[4]
    disk.set_source_path("/dev/pool-logical/diskvol1")
    disk.sync_path_props()

    _alter_compare(kvmconn, guest.get_xml(), outfile)


def testGuestBootorder():
    kvmconn = utils.URIs.open_kvm()
    guest, outfile = _get_test_content(kvmconn, "bootorder")

    assert guest.get_boot_order() == ["./devices/disk[1]"]
    assert guest.get_boot_order(legacy=True) == ["hd"]

    legacy_order = ["hd", "fd", "cdrom", "network"]
    dev_order = [
        "./devices/disk[1]",
        "./devices/disk[3]",
        "./devices/disk[2]",
        "./devices/interface[1]",
    ]
    guest.set_boot_order(legacy_order, legacy=True)
    assert guest.get_boot_order() == dev_order
    assert guest.get_boot_order(legacy=True) == legacy_order

    guest.set_boot_order(dev_order)
    assert guest.get_boot_order() == dev_order
    assert guest.get_boot_order(legacy=True) == []

    _alter_compare(kvmconn, guest.get_xml(), outfile)


def testDiskChangeBus():
    conn = utils.URIs.open_testdefault_cached()
    guest, outfile = _get_test_content(conn, "disk-change-bus")

    disk = guest.devices.disk[0]

    # Same bus is a no-op
    origxml = disk.get_xml()
    disk.change_bus(guest, "virtio")
    assert origxml == disk.get_xml()
    disk.change_bus(guest, "ide")

    disk = guest.devices.disk[2]
    disk.change_bus(guest, "scsi")

    _alter_compare(conn, guest.get_xml(), outfile)


##################
# Snapshot tests #
##################


def testChangeSnapshot():
    conn = utils.URIs.open_testdefault_cached()
    basename = "change-snapshot"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    snap = virtinst.DomainSnapshot(conn, parsexml=open(infile).read())

    check = _make_checker(snap)
    check("name", "offline-root-child1", "name-foo")
    check("state", "shutoff", "somestate")
    check("description", "offline desk", "foo\nnewline\n   indent")
    check("parent", "offline-root", "newparent")
    check("creationTime", 1375905916, 1234)
    check("memory_type", "no", "external")
    check("memory_file", None, "/some/path/to/memory.img")

    check = _make_checker(snap.disks[0])
    check("name", "hda", "hdb")
    check("snapshot", "internal", "no")

    utils.diff_compare(snap.get_xml(), outfile)


#################
# Storage tests #
#################


def testFSPool():
    conn = utils.URIs.open_testdefault_cached()
    basename = "pool-fs"
    infile = DATADIR + "%s.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    pool = virtinst.StoragePool(conn, parsexml=open(infile).read())

    check = _make_checker(pool)
    check("type", "fs", "dir")
    check("name", "pool-fs", "foo-new")
    check("uuid", "10211510-2115-1021-1510-211510211510", "10211510-2115-1021-1510-211510211999")
    check("capacity", 984373075968, 200000)
    check("allocation", 756681687040, 150000)
    check("available", 227691388928, 50000)

    check("format", "auto", "ext3")
    check("source_path", "/some/source/path", "/dev/foo/bar")
    check("target_path", "/some/target/path", "/mnt/my/foo")
    check("source_name", None, "fooname")

    utils.diff_compare(pool.get_xml(), outfile)
    utils.check_create(conn, pool.get_xml(), "storagePoolDefineXML")


def testISCSIPool():
    conn = utils.URIs.open_testdefault_cached()
    basename = "pool-iscsi"
    infile = utils.DATADIR + "/storage/%s.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    pool = virtinst.StoragePool(conn, parsexml=open(infile).read())

    check = _make_checker(pool)
    check("iqn", "foo.bar.baz.iqn", "my.iqn")
    check = _make_checker(pool.hosts[0])
    check("name", "some.random.hostname", "my.host")

    utils.diff_compare(pool.get_xml(), outfile)
    utils.check_create(conn, pool.get_xml(), "storagePoolDefineXML")


def testGlusterPool():
    conn = utils.URIs.open_testdefault_cached()
    basename = "pool-gluster"
    infile = utils.DATADIR + "/storage/%s.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    pool = virtinst.StoragePool(conn, parsexml=open(infile).read())

    check = _make_checker(pool)
    check("source_path", "/some/source/path", "/foo")
    check = _make_checker(pool.hosts[0])
    check("name", "some.random.hostname", "my.host")

    utils.diff_compare(pool.get_xml(), outfile)
    utils.check_create(conn, pool.get_xml(), "storagePoolDefineXML")


def testRBDPool():
    conn = utils.URIs.open_testdefault_cached()
    basename = "pool-rbd"
    infile = DATADIR + "%s.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    pool = virtinst.StoragePool(conn, parsexml=open(infile).read())

    check = _make_checker(pool.hosts[0])
    check("name", "ceph-mon-1.example.com")
    check("port", 6789, 1234)
    check = _make_checker(pool.hosts[1])
    check("name", "ceph-mon-2.example.com", "foo.bar")
    check("port", 6789)
    check = _make_checker(pool.hosts[2])
    check("name", "ceph-mon-3.example.com")
    check("port", 6789, 1000)
    hostobj = pool.hosts.add_new()
    hostobj.name = "frobber"
    hostobj.port = "5555"

    utils.diff_compare(pool.get_xml(), outfile)
    utils.check_create(conn, pool.get_xml(), "storagePoolDefineXML")


def testVol():
    conn = utils.URIs.open_testdefault_cached()
    basename = "pool-dir-vol"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    vol = virtinst.StorageVolume(conn, parsexml=open(infile).read())

    check = _make_checker(vol)
    check("type", None, "file")
    check("key", None, "fookey")
    check("capacity", 10737418240, 2000)
    check("allocation", 5368709120, 1000)
    check("format", "raw", "qcow2")
    check("target_path", None, "/foo/bar")
    check("backing_store", "/foo/bar/baz", "/my/backing")
    check("lazy_refcounts", False, True)

    check = _make_checker(vol.permissions)
    check("mode", "0700", "0744")
    check("owner", "10736", "10000")
    check("group", "10736", "10000")
    check("label", None, "foo.label")

    utils.diff_compare(vol.get_xml(), outfile)


###################
# <network> tests #
###################


def testNetMulti():
    conn = utils.URIs.open_testdefault_cached()
    basename = "network-multi"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    net = virtinst.Network(conn, parsexml=open(infile).read())

    check = _make_checker(net)
    check("name", "ipv6_multirange", "new-foo")
    check("uuid", "41b4afe4-87bb-8087-6724-5e208a2d483a", "41b4afe4-87bb-8087-6724-5e208a2d1111")
    check("bridge", "virbr3", "virbr3new")
    check("stp", True, False)
    check("delay", 0, 2)
    check("domain_name", "net7", "newdom")
    check("ipv6", None, True)
    check("macaddr", None, "52:54:00:69:eb:FF")
    check("virtualport_type", None, "openvswitch")

    assert len(net.portgroups) == 2
    check = _make_checker(net.portgroups[0])
    check("name", "engineering", "foo")
    check("default", True, False)

    assert len(net.ips) == 4
    check = _make_checker(net.ips[0])
    check("address", "192.168.7.1", "192.168.8.1")
    check("netmask", "255.255.255.0", "255.255.254.0")
    assert net.can_pxe() is False
    check("tftp", None, "/var/lib/tftproot")
    check("bootp_file", None, "pxeboot.img")
    check("bootp_server", None, "1.2.3.4")
    assert net.can_pxe() is True

    check = _make_checker(net.forward)
    check("mode", "nat", "route")
    check("dev", None, "eth22")
    assert net.can_pxe() is True

    check = _make_checker(net.ips[0].ranges[0])
    check("start", "192.168.7.128", "192.168.8.128")
    check("end", "192.168.7.254", "192.168.8.254")

    check = _make_checker(net.ips[0].hosts[1])
    check("macaddr", "52:54:00:69:eb:91", "52:54:00:69:eb:92")
    check("name", "badbob", "newname")
    check("ip", "192.168.7.3", "192.168.8.3")

    check = _make_checker(net.ips[1])
    check("family", "ipv6", "ipv6")
    check("prefix", 64, 63)

    r = net.routes.add_new()
    r.family = "ipv4"
    r.address = "192.168.8.0"
    r.prefix = "24"
    r.gateway = "192.168.8.10"
    check = _make_checker(r)
    check("netmask", None, "foo", None)

    utils.diff_compare(net.get_xml(), outfile)
    utils.check_create(conn, net.get_xml(), "networkDefineXML")


def testNetOpen():
    conn = utils.URIs.open_testdefault_cached()
    basename = "network-open"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    net = virtinst.Network(conn, parsexml=open(infile).read())

    check = _make_checker(net)
    check("name", "open", "new-foo")
    check("domain_name", "open", "newdom")

    check = _make_checker(net.forward)
    check("mode", "open")
    check("dev", None)

    assert len(net.ips) == 1
    check = _make_checker(net.ips[0])
    check("address", "192.168.100.1", "192.168.101.1")
    check("netmask", "255.255.255.0", "255.255.254.0")

    check = _make_checker(net.ips[0].ranges[0])
    check("start", "192.168.100.128", "192.168.101.128")
    check("end", "192.168.100.254", "192.168.101.254")

    utils.diff_compare(net.get_xml(), outfile)
    utils.check_create(conn, net.get_xml(), "networkDefineXML")


def testNetVfPool():
    conn = utils.URIs.open_testdefault_cached()
    basename = "network-vf-pool"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    net = virtinst.Network(conn, parsexml=open(infile).read())

    check = _make_checker(net)
    check("name", "passthrough", "new-foo")

    check = _make_checker(net.forward)
    check("mode", "hostdev")
    check("managed", "yes")

    check = _make_checker(net.forward.pf[0])
    check("dev", "eth3")

    utils.diff_compare(net.get_xml(), outfile)
    utils.check_create(conn, net.get_xml(), "networkDefineXML")


##############
# Misc tests #
##############


def testCPUUnknownClear():
    # Make sure .clear() even removes XML elements we don't know about
    kvmconn = utils.URIs.open_kvm()
    basename = "clear-cpu-unknown-vals"
    infile = DATADIR + "%s-in.xml" % basename
    outfile = DATADIR + "%s-out.xml" % basename
    guest = virtinst.Guest(kvmconn, parsexml=open(infile).read())

    guest.cpu.set_special_mode(guest, "host-model-only")
    guest.cpu.clear()
    utils.diff_compare(guest.get_xml(), outfile)


def testDomainRoundtrip():
    conn = utils.URIs.open_testdefault_cached()
    # Make sure our XML engine doesn't mangle non-libvirt XML bits
    infile = DATADIR + "domain-roundtrip.xml"
    outfile = DATADIR + "domain-roundtrip.xml"
    guest = virtinst.Guest(conn, parsexml=open(infile).read())

    utils.diff_compare(guest.get_xml(), outfile)


def testYesNoUnexpectedParse():
    conn = utils.URIs.open_testdefault_cached()
    # Make sure that if we see an unexpected yes/no or on/off value,
    # we just return it to the user and don't error. Libvirt could
    # change our assumptions and we shouldn't be too restrictive
    xml = (
        "<hostdev managed='foo'>\n  <rom bar='wibble'/>\n"
        "  <source><address bus='hello'/></source>\n</hostdev>"
    )
    dev = virtinst.DeviceHostdev(conn, parsexml=xml)

    assert dev.managed == "foo"
    assert dev.rom_bar == "wibble"
    assert dev.scsi_bus == "hello"

    dev.managed = "test1"
    dev.rom_bar = "test2"
    assert dev.managed == "test1"
    assert dev.rom_bar == "test2"

    with pytest.raises(ValueError):
        dev.scsi_bus = "goodbye"


def testXMLBuilderCoverage():
    """
    Test XMLBuilder corner cases
    """
    conn = utils.URIs.open_testdefault_cached()

    with pytest.raises(RuntimeError, match=".*'foo'.*"):
        # Ensure we validate root element
        virtinst.DeviceDisk(conn, parsexml="<foo/>")

    with pytest.raises(Exception, match=".*xmlParseDoc.*"):
        # Ensure we validate root element
        virtinst.DeviceDisk(conn, parsexml=-1)

    with pytest.raises(virtinst.xmlutil.DevError):
        raise virtinst.xmlutil.DevError("for coverage")

    with pytest.raises(ValueError):
        virtinst.DeviceDisk.validate_generic_name("objtype", None)

    with pytest.raises(ValueError):
        virtinst.DeviceDisk.validate_generic_name("objtype", "foo/bar")

    assert virtinst.DeviceDisk.validate_generic_name("objtype", "foo bar") is None

    # Test property __repr__ for code coverage
    assert "DeviceAddress" in str(virtinst.DeviceDisk.address)
    assert "./driver/@cache" in str(virtinst.DeviceDisk.driver_cache)

    # Conversion of 0x value into int
    xml = """
        <controller type='scsi'>
          <address type='pci' bus='0x00' slot='0x04' function='0x7'/>
        </controller>
    """
    dev = virtinst.DeviceController(conn, parsexml=xml)
    assert dev.address.slot == 4

    # Some XML formatting and get_xml_* corner cases
    conn = utils.URIs.openconn(utils.URIs.test_suite)
    xml = conn.lookupByName("test-for-virtxml").XMLDesc(0)
    guest = virtinst.Guest(conn, parsexml=xml)

    assert guest.features.get_xml().startswith("  <features")
    assert guest.clock.get_xml().startswith("  <clock")
    assert guest.seclabels[0].get_xml().startswith("<seclabel")
    assert guest.cpu.get_xml().startswith("  <cpu")
    assert guest.os.get_xml().startswith("  <os")
    assert guest.cpu.get_xml_id() == "./cpu"
    assert guest.cpu.get_xml_idx() == 0
    assert guest.get_xml_id() == "."
    assert guest.get_xml_idx() == 0

    assert guest.devices.disk[1].get_xml_id() == "./devices/disk[2]"
    assert guest.devices.disk[1].get_xml_idx() == 1


def testReplaceChildParse():
    conn = utils.URIs.open_testdefault_cached()
    buildfile = DATADIR + "replace-child-build.xml"
    parsefile = DATADIR + "replace-child-parse.xml"

    def mkdisk(target):
        disk = virtinst.DeviceDisk(conn)
        disk.device = "cdrom"
        disk.bus = "scsi"
        disk.target = target
        return disk

    guest = virtinst.Guest(conn)
    guest.add_device(mkdisk("sda"))
    guest.add_device(mkdisk("sdb"))
    guest.add_device(mkdisk("sdc"))
    guest.add_device(mkdisk("sdd"))
    guest.add_device(mkdisk("sde"))
    guest.add_device(mkdisk("sdf"))
    guest.devices.replace_child(guest.devices.disk[2], mkdisk("sdz"))
    guest.set_defaults(guest)
    utils.diff_compare(guest.get_xml(), buildfile)

    guest = virtinst.Guest(conn, parsexml=guest.get_xml())
    newdisk = virtinst.DeviceDisk(conn, parsexml=mkdisk("sdw").get_xml())
    guest.devices.replace_child(guest.devices.disk[4], newdisk)
    utils.diff_compare(guest.get_xml(), parsefile)


def testGuestXMLDeviceMatch():
    """
    Test Guest.find_device and Device.compare_device
    """
    uri = utils.URIs.test_suite
    conn = utils.URIs.openconn(uri)
    dom = conn.lookupByName("test-for-virtxml")
    xml = dom.XMLDesc(0)
    guest = virtinst.Guest(conn, xml)
    guest2 = virtinst.Guest(conn, xml)

    # Assert id matching works
    diskdev = guest.devices.disk[0]
    assert guest.find_device(diskdev) == diskdev

    # Assert type checking correct returns False
    ifacedev = guest.devices.interface[0]
    assert ifacedev.compare_device(diskdev, 0) is False

    # find_device should fail here
    nodev = virtinst.DeviceWatchdog(conn)
    assert guest.find_device(nodev) is None

    # Ensure parsed XML devices match correctly
    for srcdev in guest.devices.get_all():
        devxml = srcdev.get_xml()
        newdev = srcdev.__class__(conn, devxml)
        if srcdev != guest.find_device(newdev):
            raise AssertionError("guest.find_device failed for dev=%s" % newdev)

    # Ensure devices from another parsed XML doc compare correctly
    for srcdev in guest.devices.get_all():
        if not guest2.find_device(srcdev):
            raise AssertionError("guest.find_device failed for dev=%s" % srcdev)


def testControllerAttachedDevices():
    """
    Test DeviceController.get_attached_devices
    """
    conn = utils.URIs.open_testdefault_cached()
    xml = open(DATADIR + "controller-attached-devices.xml").read()
    guest = virtinst.Guest(conn, xml)

    # virtio-serial path
    controller = [c for c in guest.devices.controller if c.type == "virtio-serial"][0]
    devs = controller.get_attached_devices(guest)
    assert len(devs) == 4
    assert devs[-1].DEVICE_TYPE == "console"

    # disk path
    controller = [c for c in guest.devices.controller if c.type == "sata"][0]
    devs = controller.get_attached_devices(guest)
    assert len(devs) == 1
    assert devs[-1].device == "cdrom"

    # Little test for DeviceAddress.pretty_desc
    assert devs[-1].address.pretty_desc() == "0:0:0:3"


def testRefreshMachineType():
    guest = virtinst.Guest(utils.URIs.openconn(utils.URIs.kvm_x86))
    guest.os.machine = "pc-i440fx-5.2"
    guest.refresh_machine_type()
    assert guest.os.machine == "pc"

    guest = virtinst.Guest(utils.URIs.openconn(utils.URIs.kvm_x86))
    guest.os.machine = "pc-q35-XYZ"
    guest.refresh_machine_type()
    assert guest.os.machine == "q35"

    guest = virtinst.Guest(utils.URIs.openconn(utils.URIs.kvm_s390x))
    guest.os.machine = "s390-ccw-virtio-12345"
    guest.refresh_machine_type()
    assert guest.os.machine == "s390-ccw-virtio"


def testDiskSourceAbspath():
    # If an existing disk doesn't have an abspath in the XML, make sure
    # we don't convert it just by parsing
    conn = utils.URIs.open_testdefault_cached()
    xml = "<disk type='file' device='disk'><source file='foobar'/></disk>"
    disk = virtinst.DeviceDisk(conn, parsexml=xml)
    assert disk.get_source_path() == "foobar"

    # But setting a relative path should convert it
    import os

    disk.set_source_path("foobar2")
    assert disk.get_source_path() == os.path.abspath("foobar2")

    # ...unless it's a URL
    disk.set_source_path("http://example.com/foobar3")
    assert disk.get_source_path() == "http://example.com/foobar3"


def testUnknownEmulatorDomcapsLookup(monkeypatch):
    """
    Libvirt can handle defining a VM with a custom emulator, one not detected
    by `virsh capabilities`. An appropriate `virsh domcapabilities` call will
    inspect the emulator and return relevant info.

    This test ensures that for parsing XML the `virsh capabilities` failure
    isn't fatal, and we attempt to return valid `virsh domcapabilities` data
    """

    seen = False

    def fake_build_from_params(conn, emulator, arch, machine, _hvtype):
        nonlocal seen
        seen = True
        assert arch == "mips"
        assert machine == "some-unknown-machine"
        assert emulator == "/my/manual/emulator"
        return virtinst.DomainCapabilities(conn)

    monkeypatch.setattr("virtinst.DomainCapabilities.build_from_params", fake_build_from_params)

    conn = utils.URIs.open_kvm()
    xml = open(DATADIR + "emulator-custom.xml").read()
    guest = virtinst.Guest(conn, xml)
    assert guest.lookup_domcaps()
    assert guest.lookup_domcaps()
    assert seen


def testConvertToQ35():
    conn = utils.URIs.openconn(utils.URIs.kvm_x86)

    def _test(filename_base, **kwargs):
        guest, outfile = _get_test_content(conn, filename_base)
        guest.convert_to_q35(**kwargs)
        _alter_compare(conn, guest.get_xml(), outfile)

    _test("convert-to-q35-win10")
    _test("convert-to-q35-f39", num_pcie_root_ports=5)


def testConvertToVNC():
    conn = utils.URIs.openconn(utils.URIs.kvm_x86)

    def _test(filename_base, **kwargs):
        guest, outfile = _get_test_content(conn, filename_base)
        guest.convert_to_vnc(**kwargs)
        _alter_compare(conn, guest.get_xml(), outfile)

    _test("convert-to-vnc-empty", qemu_vdagent=True)
    _test("convert-to-vnc-spice-devices")
    _test("convert-to-vnc-spice-manyopts", qemu_vdagent=True)
    _test("convert-to-vnc-has-vnc", qemu_vdagent=True)



#!/usr/bin/env python3

import argparse
import os
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
from virtinst import URI


def parse_options():
    description = (
        "Generate a fake URI for use with virt-manager/virtinst "
        "that wraps a standard test:/// URI but pretends to be a different "
        "hypervisor. See virtinst/uri.py MagicURI for format details. "
        "Example: magicuri.py qemu+tcp://fakeuri.example.com/system"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("fakeuri", help="The libvirt URI we should fake")
    parser.add_argument("--capsfile", help="Path to file to use for capabilities XML")
    parser.add_argument("--domcapsfile", help="Path to file to use for domain capabilities XML")
    parser.add_argument("--driverxml", help="Path to driver xml (defaults to testdriver.xml)")

    options = parser.parse_args()

    testdir = os.path.abspath(os.path.dirname(__file__))
    capsdir = os.path.join(testdir, "data/capabilities/")

    uriobj = URI(options.fakeuri)
    hv = uriobj.scheme

    capsfile = None
    domcapsfile = None
    if hv == "qemu":
        capsfile = capsdir + "kvm-x86_64.xml"
        domcapsfile = capsdir + "kvm-x86_64-domcaps.xml"
    elif hv == "xen":
        capsfile = capsdir + "xen-rhel5.4.xml"
    elif hv == "lxc":
        capsfile = capsdir + "lxc.xml"
    elif hv == "vz":
        capsfile = capsdir + "vz.xml"

    if options.capsfile:
        capsfile = os.path.abspath(options.capsfile)
    if options.domcapsfile:
        domcapsfile = os.path.abspath(options.domcapsfile)

    driverxml = os.path.join(testdir, "testdriver.xml")
    if options.driverxml:
        driverxml = os.path.abspath(options.driverxml)

    return options.fakeuri, capsfile, domcapsfile, driverxml


def main():
    fakeuri, capsfile, domcapsfile, driverxml = parse_options()
    uri = "__virtinst_test__test://%s" % driverxml
    uri += ",fakeuri=%s" % fakeuri

    if capsfile:
        uri += ",caps=%s" % capsfile
    if domcapsfile:
        uri += ",domcaps=%s" % domcapsfile

    if driverxml and not os.path.exists(driverxml):
        print("%s does not exist" % capsfile)
        return 1
    if capsfile and not os.path.exists(capsfile):
        print("%s does not exist" % capsfile)
        return 1
    if domcapsfile and not os.path.exists(domcapsfile):
        print("%s does not exist" % domcapsfile)
        return 1

    print(uri)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#
# Common code for unattended installations
#
# Copyright 2019 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import getpass
import locale
import os
import pwd
import re
import tempfile

from gi.repository import Libosinfo

from . import urlfetcher
from .. import progress
from ..logger import log


def _is_user_login_safe(login):
    return login != "root"


def _login_from_hostuser():
    hostuser = getpass.getuser()
    realname = pwd.getpwnam(hostuser).pw_gecos
    if not _is_user_login_safe(hostuser):
        return None, None  # pragma: no cover
    return hostuser, realname  # pragma: no cover


def _make_installconfig(script, osobj, unattended_data, arch, hostname, url):
    """
    Build a Libosinfo.InstallConfig instance
    """

    def get_timezone():
        TZ_FILE = "/etc/localtime"
        linkpath = os.path.realpath(TZ_FILE)
        tokens = linkpath.split("zoneinfo/")
        if len(tokens) > 1:
            return tokens[1]

    def get_language():
        return locale.getlocale()[0]

    config = Libosinfo.InstallConfig()

    # Set user login and name
    # In case it's specified via command-line, use the specified one as login
    # and realname. Otherwise, fallback fto the one from the system
    login = unattended_data.user_login
    realname = unattended_data.user_login
    if not login:
        login, realname = _login_from_hostuser()

    if login:
        login = login.lower()
        if not _is_user_login_safe(login):
            raise RuntimeError(
                _("%(osname)s cannot use '%(loginname)s' as user-login.")
                % {"osname": osobj.name, "loginname": login}
            )

        config.set_user_login(login)
        config.set_user_realname(realname)

    # Set user-password.
    # In case it's required and not passed, just raise a RuntimeError.
    if script.requires_user_password() and not unattended_data.get_user_password():
        raise RuntimeError(_("%s requires the user-password to be set.") % osobj.name)
    config.set_user_password(unattended_data.get_user_password() or "")

    # Set the admin-password:
    # In case it's required and not passed, just raise a RuntimeError.
    if script.requires_admin_password() and not unattended_data.get_admin_password():
        raise RuntimeError(_("%s requires the admin-password to be set.") % osobj.name)
    config.set_admin_password(unattended_data.get_admin_password() or "")

    # Set the target disk.
    # virtiodisk is the preferred way, in case it's supported, otherwise
    # just fallback to scsi.
    #
    # Note: this is linux specific and will require some changes whenever
    # support for Windows will be added.
    tgt = "/dev/vda" if osobj.supports_virtiodisk() else "/dev/sda"
    if osobj.is_windows():
        tgt = "C"
    config.set_target_disk(tgt)

    # Set hardware architecture and hostname
    config.set_hardware_arch(arch)

    # Some installations will bail if the Computer's name contains one of the
    # following characters: "[{|}~[\\]^':; <=>?@!\"#$%`()+/.,*&]".
    # In order to take a safer path, let's ensure that we never set those,
    # replacing them by "-".
    hostname = re.sub("[{|}~[\\]^':; <=>?@!\"#$%`()+/.,*&]", "-", hostname)
    config.set_hostname(hostname)

    # Try to guess the timezone from '/etc/localtime', in case it's not
    # possible 'America/New_York' will be used.
    timezone = get_timezone()
    if timezone:
        config.set_l10n_timezone(timezone)

    # Try to guess to language and keyboard layout from the system's
    # language.
    #
    # This method has flaws as it's quite common to have language and
    # keyboard layout not matching. Otherwise, there's no easy way to guess
    # the keyboard layout without relying on a set of APIs of an specific
    # Desktop Environment.
    language = get_language()
    if language:
        config.set_l10n_language(language)
        config.set_l10n_keyboard(language)

    if url:
        config.set_installation_url(url)  # pylint: disable=no-member

    if unattended_data.reg_login:
        config.set_reg_login(unattended_data.reg_login)

    if unattended_data.product_key:
        config.set_reg_product_key(unattended_data.product_key)

    log.debug("InstallScriptConfig created with the following params:")
    log.debug("username: %s", config.get_user_login())
    log.debug("realname: %s", config.get_user_realname())
    log.debug("target disk: %s", config.get_target_disk())
    log.debug("hardware arch: %s", config.get_hardware_arch())
    log.debug("hostname: %s", config.get_hostname())
    log.debug("timezone: %s", config.get_l10n_timezone())
    log.debug("language: %s", config.get_l10n_language())
    log.debug("keyboard: %s", config.get_l10n_keyboard())
    if hasattr(config, "get_installation_url"):
        log.debug("url: %s", config.get_installation_url())  # pylint: disable=no-member
    log.debug("reg-login %s", config.get_reg_login())
    log.debug("product-key: %s", config.get_reg_product_key())

    return config


class OSInstallScript:
    """
    Wrapper for Libosinfo.InstallScript interactions
    """

    @staticmethod
    def have_new_libosinfo():
        from ..osdict import OSDB

        win7 = OSDB.lookup_os("win7")
        for script in win7.get_install_script_list():
            if Libosinfo.InstallScriptInjectionMethod.CDROM & script.get_injection_methods():
                return True
        return False  # pragma: no cover

    @staticmethod
    def have_libosinfo_installation_url():
        return hasattr(Libosinfo.InstallConfig, "set_installation_url")

    def __init__(self, script, osobj, osinfomediaobj, osinfotreeobj):
        self._script = script
        self._osobj = osobj
        self._osinfomediaobj = osinfomediaobj
        self._osinfotreeobj = osinfotreeobj
        self._config = None

        if not OSInstallScript.have_new_libosinfo():  # pragma: no cover
            raise RuntimeError(
                _("libosinfo or osinfo-db is too old to support unattended installs.")
            )

    def get_expected_filename(self):
        return self._script.get_expected_filename()

    def set_preferred_injection_method(self, namestr):
        # If we ever make this user configurable, this will need to be smarter
        names = {
            "cdrom": Libosinfo.InstallScriptInjectionMethod.CDROM,
            "initrd": Libosinfo.InstallScriptInjectionMethod.INITRD,
        }

        log.debug("Using '%s' injection method", namestr)
        injection_method = names[namestr]
        supported_injection_methods = self._script.get_injection_methods()
        if injection_method & supported_injection_methods == 0:
            raise RuntimeError(
                _("OS '%(osname)s' does not support required injection method '%(methodname)s'")
                % {"osname": self._osobj.name, "methodname": namestr}
            )

        self._script.set_preferred_injection_method(injection_method)

    def set_installation_source(self, namestr):
        # If we ever make this user configurable, this will need to be smarter
        names = {
            "media": Libosinfo.InstallScriptInstallationSource.MEDIA,
            "network": Libosinfo.InstallScriptInstallationSource.NETWORK,
        }

        log.debug("Using '%s' installation source", namestr)
        self._script.set_installation_source(names[namestr])

    def _requires_param(self, config_param):
        param = self._script.get_config_param(config_param)
        return bool(param and not param.is_optional())

    def requires_user_password(self):
        return self._requires_param(Libosinfo.INSTALL_CONFIG_PROP_USER_PASSWORD)

    def requires_admin_password(self):
        return self._requires_param(Libosinfo.INSTALL_CONFIG_PROP_ADMIN_PASSWORD)

    def set_config(self, config):
        self._config = config

    def generate(self):
        if self._osinfomediaobj:
            return self._script.generate_for_media(self._osinfomediaobj, self._config)
        if hasattr(self._script, "generate_for_tree") and self._osinfotreeobj:
            # osinfo_install_script_generate_for_tree() is part of
            # libosinfo 1.6.0
            return self._script.generate_for_tree(self._osinfotreeobj, self._config)

        return self._script.generate(self._osobj.get_handle(), self._config)

    def generate_cmdline(self):
        if self._osinfomediaobj:
            return self._script.generate_command_line_for_media(self._osinfomediaobj, self._config)
        if hasattr(self._script, "generate_command_line_for_tree") and self._osinfotreeobj:
            # osinfo_install_script_generate_command_line_for_tree() is part of
            # libosinfo 1.6.0
            return self._script.generate_command_line_for_tree(self._osinfotreeobj, self._config)
        return self._script.generate_command_line(self._osobj.get_handle(), self._config)

    def _generate_debug(self):
        original_user_password = self._config.get_user_password()
        original_admin_password = self._config.get_admin_password()

        self._config.set_user_password("[SCRUBBLED]")
        self._config.set_admin_password("[SCRUBBLED]")

        debug_content = self.generate()

        self._config.set_user_password(original_user_password)
        self._config.set_admin_password(original_admin_password)

        return debug_content

    def write(self):
        fileobj = tempfile.NamedTemporaryFile(prefix="virtinst-unattended-script", delete=False)
        scriptpath = fileobj.name

        content = self.generate()
        open(scriptpath, "w").write(content)

        debug_content = self._generate_debug()

        log.debug("Generated unattended script: %s", scriptpath)
        log.debug("Generated script contents:\n%s", debug_content)

        return scriptpath


class UnattendedData:
    profile = None
    admin_password_file = None
    user_login = None
    user_password_file = None
    product_key = None
    reg_login = None

    def _get_password(self, pwdfile):
        with open(pwdfile, "r") as fobj:
            return fobj.readline().rstrip("\n\r")

    def get_user_password(self):
        if self.user_password_file:
            return self._get_password(self.user_password_file)

    def get_admin_password(self):
        if self.admin_password_file:
            return self._get_password(self.admin_password_file)


def _make_scriptmap(script_list):
    """
    Generate a mapping of profile name -> [list, of, rawscripts]
    """
    script_map = {}
    for script in script_list:
        profile = script.get_profile()
        if profile not in script_map:
            script_map[profile] = []
        script_map[profile].append(script)
    return script_map


def _find_default_profile(profile_names):
    profile_prefs = ["desktop"]
    found = None
    for p in profile_prefs:
        if p in profile_names:
            found = p
            break
    return found or profile_names[0]


def _lookup_rawscripts(osinfo, profile, os_media):
    script_list = []

    if os_media:
        if not os_media.supports_installer_script():
            # This is a specific annotation for media like livecds that
            # don't support unattended installs
            raise RuntimeError(
                _("OS '%s' media does not support unattended installation") % (osinfo.name)
            )

        # In case we're dealing with a media installation, let's try to get
        # the installer scripts from the media, in case any is set.
        script_list = os_media.get_install_script_list()

    if not script_list:
        script_list = osinfo.get_install_script_list()
    if not script_list:
        raise RuntimeError(_("OS '%s' does not support unattended installation.") % osinfo.name)

    script_map = _make_scriptmap(script_list)
    profile_names = list(sorted(script_map.keys()))
    if profile:
        rawscripts = script_map.get(profile, [])
        if not rawscripts:
            raise RuntimeError(
                _(
                    "OS '%(osname)s' does not support unattended "
                    "installation for the '%(profilename)s' profile. "
                    "Available profiles: %(profiles)s"
                )
                % {
                    "osname": osinfo.name,
                    "profilename": profile,
                    "profiles": ", ".join(profile_names),
                }
            )
    else:
        profile = _find_default_profile(profile_names)
        log.warning(_("Using unattended profile '%s'"), profile)
        rawscripts = script_map[profile]

    # Some OSes (as Windows) have more than one installer script,
    # depending on the OS version and profile chosen, to be used to
    # perform the unattended installation.
    ids = []
    for rawscript in rawscripts:
        ids.append(rawscript.get_id())

    log.debug("Install scripts found for profile '%s': %s", profile, ", ".join(ids))
    return rawscripts


def prepare_install_scripts(guest, unattended_data, url, os_media, os_tree, injection_method):
    def _get_installation_source(os_media):
        if not os_media:
            return "network"
        return "media"

    scripts = []
    rawscripts = _lookup_rawscripts(guest.osinfo, unattended_data.profile, os_media)

    osinfomediaobj = os_media.get_osinfo_media() if os_media else None
    osinfotreeobj = os_tree.get_osinfo_tree() if os_tree else None

    for rawscript in rawscripts:
        script = OSInstallScript(rawscript, guest.osinfo, osinfomediaobj, osinfotreeobj)

        script.set_preferred_injection_method(injection_method)

        installationsource = _get_installation_source(os_media)
        script.set_installation_source(installationsource)

        config = _make_installconfig(
            script, guest.osinfo, unattended_data, guest.os.arch, guest.name, url
        )
        script.set_config(config)
        scripts.append(script)
    return scripts


def download_drivers(locations, scratchdir, meter):
    meter = progress.ensure_meter(meter)
    fetcher = urlfetcher.DirectFetcher(None, scratchdir, meter)
    fetcher.meter = meter

    drivers = []

    try:
        for location in locations:
            filename = location.rsplit("/", 1)[1]
            driver = fetcher.acquireFile(location)
            drivers.append((driver, filename))
    except Exception:  # pragma: no cover
        for driverpair in drivers:
            os.unlink(driverpair[0])
        raise

    return drivers



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


############################################
# UI tests for virt-manager's VM snapshots #
############################################

_DEFAULT_VMNAME = "test-snapshots"


def _open_snapshots_window(app, vmname=_DEFAULT_VMNAME):
    if not app.is_running():
        app.open(show_console=vmname)
    win = app.find_details_window(vmname)
    win.find("Snapshots", "radio button").click()
    return win


def testSnapshotsSmokeTest(app):
    """
    Smoke test to ensure all snapshots show correctly
    """
    win = _open_snapshots_window(app)
    errlabel = win.find("snapshot-error-label", "label")
    lst = win.find("snapshot-list", "table")
    lib.utils.walkUIList(app, win, lst, lambda: errlabel.showing)


def testSnapshotLifecycle(app):
    """
    Create/delete/start/stop snapshots
    """
    win = _open_snapshots_window(app)
    vmrun = win.find("Run", "push button")
    vmpause = win.find("Pause", "toggle button")
    snaprun = win.find("snapshot-start", "push button")

    # Start already running snapshot
    snapname = "internal-root"
    win.find(snapname, "table cell").click()
    snaprun.click()
    app.click_alert_button("run the snapshot '%s'" % snapname, "Yes")
    lib.utils.check(lambda: not vmrun.sensitive)

    # Start offline snapshot
    snapname = "offline-root"
    win.find(snapname, "table cell").click()
    snaprun.click()
    app.click_alert_button("run the snapshot '%s'" % snapname, "No")
    lib.utils.check(lambda: not vmrun.sensitive)
    snaprun.click()
    app.click_alert_button("run the snapshot '%s'" % snapname, "Yes")
    lib.utils.check(lambda: vmrun.sensitive)

    # Start paused snapshot
    snapname = "snap-paused"
    win.find(snapname, "table cell").click()
    snaprun.click()
    app.click_alert_button("run the snapshot '%s'" % snapname, "Yes")
    lib.utils.check(lambda: vmpause.checked)

    # Edit snapshot
    descui = win.find("snapshot-description")
    desc = "TESTSNAP"
    descui.set_text(desc)
    win.find("snapshot-apply", "push button").click()
    win.find("snapshot-refresh", "push button").click()
    lib.utils.check(lambda: descui.text == desc)
    # Apply by clicking away
    desc += " ROUND2"
    descui.set_text(desc)
    win.find("internal-root", "table cell").click()
    app.click_alert_button("There are unapplied changes", "Yes")

    # Create new snapshot
    win.find("snapshot-add", "push button").click()
    newwin = app.find_window("Create snapshot")
    snapname = "testnewsnap"
    newwin.find("Name:", "text").set_text(snapname)
    newwin.find("internal", "radio button").click()
    newwin.find("Description:", "text").set_text("testdesc")
    newwin.find("Finish", "push button").click()
    lib.utils.check(lambda: not newwin.showing)
    newc = win.find(snapname, "table cell")
    lib.utils.check(lambda: newc.state_selected)

    # Delete it
    win.find("snapshot-delete", "push button").click()
    app.click_alert_button("permanently delete", "No")
    lib.utils.check(lambda: not newc.dead)
    win.find("snapshot-delete", "push button").click()
    app.click_alert_button("permanently delete", "Yes")
    lib.utils.check(lambda: newc.dead)

    # Recreate another snapshot with the same name
    win.find("snapshot-add", "push button").click()
    newwin = app.find_window("Create snapshot")
    snapname = "testnewsnap"
    newwin.find("Name:", "text").set_text(snapname)
    newwin.find("internal", "radio button").click()
    newwin.find("Finish", "push button").click()
    lib.utils.check(lambda: not newwin.showing)
    newc = win.find(snapname, "table cell")
    lib.utils.check(lambda: newc.state_selected)

    # Switch out of window
    win.find("Details", "radio button").click()
    lib.utils.check(lambda: not snaprun.showing)


def testSnapshotMisc1(app):
    """
    Test snapshot corner cases
    """
    vmname = "test-state-managedsave"
    win = _open_snapshots_window(app, vmname)
    vmrun = win.find("Restore", "push button")

    # Create new snapshot
    win.find("snapshot-add", "push button").click()
    app.click_alert_button("not become part of the snapshot", "Cancel")
    lib.utils.check(lambda: win.active)
    win.find("snapshot-add", "push button").click()
    app.click_alert_button("not become part of the snapshot", "OK")
    newwin = app.find_window("Create snapshot")
    snapname1 = "testnewsnap1"
    newwin.find("Name:", "text").set_text(snapname1)
    newwin.find("internal", "radio button").click()
    newwin.find("Finish", "push button").click()
    lib.utils.check(lambda: not newwin.showing)
    newc = win.find(snapname1, "table cell")
    lib.utils.check(lambda: newc.state_selected)

    # Start the VM, create another snapshot
    vmrun.click()
    lib.utils.check(lambda: not vmrun.sensitive)
    win.find("snapshot-add", "push button").click()
    newwin = app.find_window("Create snapshot")
    newwin.find("internal", "radio button").click()
    # Force validation error
    newwin.find("Name:", "text").set_text("bad name")
    newwin.find("Finish", "push button").click()
    app.click_alert_button("validating snapshot", "OK")
    # Force name collision
    newwin.find("Name:", "text").set_text(snapname1)
    newwin.find("Finish", "push button").click()
    app.click_alert_button(snapname1, "Close")
    # Make it succeed
    snapname2 = "testnewsnap2"
    newwin.find("Name:", "text").set_text(snapname2)
    newwin.find("Finish", "push button").click()
    lib.utils.check(lambda: not newwin.showing)
    newc = win.find(snapname2, "table cell")
    lib.utils.check(lambda: newc.state_selected)

    # Trigger another managed save warning
    smenu = win.find("Menu", "toggle button")
    smenu.click()
    save = smenu.find("Save", "menu item")
    save.click()
    lib.utils.check(lambda: vmrun.sensitive)
    win.find(snapname1, "table cell").click(button=3)
    app.root.find("Start snapshot", "menu item").click()
    app.click_alert_button("run the snapshot '%s'" % snapname1, "Yes")
    app.click_alert_button("no memory state", "Cancel")
    win.find("snapshot-start").click()
    app.click_alert_button("run the snapshot '%s'" % snapname1, "Yes")
    app.click_alert_button("no memory state", "OK")

    # Multi select
    cell1 = win.find(snapname1, "table cell")
    cell2 = win.find(snapname2, "table cell")
    cell1.click()
    app.rawinput.holdKey("Shift_L")
    app.rawinput.pressKey("Down")
    app.rawinput.releaseKey("Shift_L")
    win.find("snapshot-delete").click()
    app.click_alert_button("permanently delete", "Yes")
    lib.utils.check(lambda: cell1.dead)
    lib.utils.check(lambda: cell2.dead)
    lib.utils.check(lambda: win.active)



# Copyright 2019 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import logging

# This is exported by virtinst/__init__.py
log = logging.getLogger("virtinst")


def reset_logging():
    rootLogger = logging.getLogger()

    # Undo early logging
    for handler in rootLogger.handlers:
        rootLogger.removeHandler(handler)

    # Undo any logging on our log handler. Needed for test suite
    for handler in log.handlers:
        log.removeHandler(handler)



# Copyright (C) 2013, 2015 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import tempfile

from tests import utils

from virtinst import Cloner


CLI_XMLDIR = utils.DATADIR + "/cli/virtclone/"


def test_clone_unmanaged():
    """
    Test that unmanaged storage duplication via the clone wizard
    actually copies data
    """
    xmlpath = CLI_XMLDIR + "clone-disk.xml"
    conn = utils.URIs.open_testdefault_cached()
    xml = open(xmlpath).read()

    tmp1 = tempfile.NamedTemporaryFile()
    tmp2 = tempfile.NamedTemporaryFile()
    inp1 = os.path.abspath(__file__)
    inp2 = xmlpath

    xml = xml.replace("/tmp/__virtinst_cli_exist1.img", inp1)
    xml = xml.replace("/tmp/__virtinst_cli_exist2.img", inp2)
    cloner = Cloner(conn, src_xml=xml)

    diskinfos = cloner.get_nonshare_diskinfos()
    assert len(diskinfos) == 2
    diskinfos[0].set_new_path(tmp1.name, False)
    diskinfos[1].set_new_path(tmp2.name, False)

    cloner.prepare()
    cloner.start_duplicate(None)

    assert open(tmp1.name).read() == open(inp1).read()
    assert open(tmp2.name).read() == open(inp2).read()


def test_generate_name():
    conn = utils.URIs.open_testdriver_cached()

    def _g(n):
        return Cloner.generate_clone_name(conn, n)

    assert _g("test") == "test-clone1"
    assert _g("test-clone-simple") == "test-clone-simple-clone"
    assert _g("test-clone-simple-clone") == "test-clone-simple-clone1"
    assert _g("test-clone-simple-clone5") == "test-clone-simple-clone6"



#
# Copyright 2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceVideo(Device):
    XML_NAME = "video"

    _XML_PROP_ORDER = ["_model", "vram", "heads", "vgamem"]
    vram = XMLProperty("./model/@vram", is_int=True)
    vram64 = XMLProperty("./model/@vram64", is_int=True)
    ram = XMLProperty("./model/@ram", is_int=True)
    heads = XMLProperty("./model/@heads", is_int=True)
    vgamem = XMLProperty("./model/@vgamem", is_int=True)
    accel3d = XMLProperty("./model/acceleration/@accel3d", is_yesno=True)
    blob = XMLProperty("./model/@blob", is_onoff=True)
    primary = XMLProperty("./model/@primary", is_yesno=True)

    def _set_model(self, val):
        self._model = val
        if self._model != "qxl":
            self.ram = None
            self.vgamem = None
            self.vram64 = None

    def _get_model(self):
        return self._model

    _model = XMLProperty("./model/@type")
    model = property(_get_model, _set_model)

    ##################
    # Default config #
    ##################

    @staticmethod
    def default_model(guest):
        if not guest.os.is_hvm():
            return None
        if guest.os.is_pseries():
            return "vga"
        if guest.os.is_loongarch64():
            return "virtio"
        if guest.os.is_riscv_virt():
            # For all cases here the hv and guest are new enough for virtio
            return "virtio"
        if guest.os.is_s390x() and guest.conn.is_qemu():
            # s390x doesn't support any of the PCI video devices
            return "virtio"
        if guest.has_spice() and guest.has_gl():
            # virtio is implied in this case
            return "virtio"

        if guest.lookup_domcaps().supports_video_virtio() and guest.osinfo.supports_virtiogpu():
            # When the guest supports it, this is the top preference
            return "virtio"
        if guest.os.is_x86() and guest.has_spice() and guest.lookup_domcaps().supports_video_qxl():
            # qxl is only beneficial over regular vga when paired with spice.
            # The device still may not be available though
            return "qxl"
        if guest.is_uefi():
            if guest.os.is_arm():
                return "ramfb"
            if guest.lookup_domcaps().supports_video_bochs():
                return "bochs"
        return "vga"

    def set_defaults(self, guest):
        if not self.model:
            self.model = self.default_model(guest)
        if self.model == "virtio" and guest.has_gl():
            self.accel3d = True



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import tests.utils
from . import lib


#############################################################
# UI tests for manager window, and basic VM lifecycle stuff #
#############################################################


def _testVMLifecycle(app):
    """
    Basic VM lifecycle test, shared between standard and no-events
    testing
    """
    manager = app.topwin
    shutdown = manager.find("Shut Down", "push button")
    pause = manager.find("Pause", "toggle button")
    run = manager.find("Run", "push button")
    force = manager.find("Force Off", "menu item")
    smenu = manager.find("Menu", "toggle button")
    save = manager.find("Save", "menu item")

    c = manager.find("test-many-devices", "table cell")
    c.click()
    smenu.click()
    force.click()
    app.click_alert_button("Are you sure you want", "Yes")
    lib.utils.check(lambda: run.sensitive, timeout=5)

    run.click()
    lib.utils.check(lambda: not run.sensitive, timeout=5)
    pause.click()
    lib.utils.check(lambda: pause.checked, timeout=5)
    pause.click()
    lib.utils.check(lambda: not pause.checked, timeout=5)
    smenu.click()
    save.click()
    lib.utils.check(lambda: run.sensitive, timeout=5)
    lib.utils.check(lambda: "Saved" in c.text)
    run.click()
    lib.utils.check(lambda: shutdown.sensitive, timeout=5)


def testVMLifecycle(app):
    # qemu hits some different domain code paths for setTime
    app.uri = tests.utils.URIs.kvm_x86
    _testVMLifecycle(app)


def testVMNoEventsLifecycle(app):
    app.open(extra_opts=["--test-options=no-events", "--test-options=short-poll"])
    _testVMLifecycle(app)


def testVMLifecycleExtra(app):
    """
    Test vmmenu lifecycle options
    """
    app.open(keyfile="confirm-all.ini")
    manager = app.topwin
    run = manager.find("Run", "push button")
    shutdown = manager.find("Shut Down", "push button")
    pause = manager.find("Pause", "toggle button")

    def confirm_is_running():
        lib.utils.check(lambda: not run.sensitive)

    def confirm_is_shutdown():
        lib.utils.check(lambda: not shutdown.sensitive)

    def confirm_is_paused():
        lib.utils.check(lambda: pause.checked)

    def confirm_not_paused():
        lib.utils.check(lambda: not pause.checked)

    def test_action(**kwargs):
        app.manager_vm_action("test", confirm_click_no=True, **kwargs)

    confirm_is_running()
    test_action(reset=True)
    confirm_is_running()
    test_action(reboot=True)
    confirm_is_running()
    test_action(shutdown=True)
    confirm_is_shutdown()
    test_action(run=True)
    confirm_is_running()
    test_action(destroy=True)
    confirm_is_shutdown()
    test_action(run=True)
    confirm_is_running()
    test_action(pause=True)
    confirm_is_paused()
    test_action(resume=True)
    confirm_not_paused()
    test_action(save=True)
    confirm_is_shutdown()
    test_action(restore=True)
    confirm_is_running()


def testManagerSaveCancelError(app):
    """
    Test managed save special behavior
    """
    app.open(extra_opts=["--test-options=test-managed-save"])

    manager = app.topwin
    run = manager.find("Run", "push button")
    smenu = manager.find("Menu", "toggle button")
    save = manager.find("Save", "menu item")

    c = manager.find("test-many-devices", "table cell")
    c.click()

    # Save it, attempt a cancel operation
    smenu.click()
    save.click()
    progwin = app.find_window("Saving Virtual Machine")
    # Attempt cancel which will fail, then find the error message
    progwin.find("Cancel", "push button").click()
    progwin.find("Error cancelling save job")
    lib.utils.check(lambda: not progwin.showing, timeout=5)
    lib.utils.check(lambda: run.sensitive)

    # Restore will fail and offer to remove managed save
    run.click()
    app.click_alert_button("remove the saved state", "No")
    lib.utils.check(lambda: run.sensitive)
    run.click()
    app.click_alert_button("remove the saved state", "Yes")
    lib.utils.check(lambda: not run.sensitive)


def testManagerQEMUSetTime(app):
    """
    Fake qemu setTime behavior for code coverage
    """
    app.uri = tests.utils.URIs.kvm_x86
    manager = app.topwin
    run = manager.find("Run", "push button")
    smenu = manager.find("Menu", "toggle button")
    save = manager.find("Save", "menu item")

    c = manager.find("test alternate devs title", "table cell")
    c.click()

    # Save -> resume -> save
    smenu.click()
    save.click()
    lib.utils.check(lambda: run.sensitive)
    app.sleep(1)  # give settime thread time to run
    run.click()
    lib.utils.check(lambda: not run.sensitive)
    app.sleep(1)  # give settime thread time to run
    smenu.click()
    save.click()
    lib.utils.check(lambda: run.sensitive)
    app.sleep(1)  # give settime thread time to run


def testManagerVMRunFail(app):
    # Force VM startup to fail so we can test the error path
    app.open(extra_opts=["--test-options=test-vm-run-fail"])

    manager = app.topwin

    c = manager.find("test-clone-simple", "table cell")
    c.click()
    manager.find("Run", "push button").click()
    app.click_alert_button("fake error", "Close")


def testManagerColumns(app):
    # Enable all stat options
    # Need to expand the window size so all columns are onscreen
    app.open(keyfile="winsize.ini")
    app.root.find("Edit", "menu").click()
    app.root.find("Preferences", "menu item").click()
    win = app.find_window("Preferences")
    win.find("Polling", "page tab").click()
    win.find_fuzzy("Poll Disk", "check").click()
    win.find_fuzzy("Poll Network", "check").click()
    win.find_fuzzy("Poll Memory", "check").click()
    win.find("Close", "push button").click()

    manager = app.topwin

    def _test_sort(name):
        col = manager.find(name, "table column header")
        col.check_onscreen()
        # Trigger sorting
        col.click()
        col.click()

    def _click_column_menu(name):
        manager.find("View", "menu").click()
        menu = manager.find("Graph", "menu")
        menu.point()
        menu.find_fuzzy(name, "check menu item").click()

    def _test_column(name):
        _click_column_menu(name)
        _test_sort(name)

    _test_sort("Name")
    _click_column_menu("Guest CPU")
    _click_column_menu("Guest CPU")
    _test_sort("CPU usage")
    _test_column("Host CPU")
    _test_column("Memory")
    _test_column("Disk I/O")
    _test_column("Network I/O")


def testManagerWindowReposition(app):
    """
    Restore previous position when window is reopened
    """
    manager = app.topwin
    host = app.manager_open_host("Storage")
    fmenu = host.find("File", "menu")
    fmenu.click()
    fmenu.find("View Manager", "menu item").click()
    lib.utils.check(lambda: manager.active)

    # Use alt+f7 combo to move window
    curxy = manager.title_coordinates()
    newxy = (curxy[0] + 400, curxy[1] + 400)
    manager.keyCombo("<alt>F7")
    app.rawinput.click(*newxy)
    checkxy = manager.position[0], manager.position[1]
    manager.window_close()
    host.click_title()
    host.find("File", "menu").click()
    host.find("View Manager", "menu item").click()
    lib.utils.check(lambda: manager.showing)

    # Results can be off by one or two, but it's not a virt-manager bug
    assert abs(manager.position[0] - checkxy[0]) in range(3)
    assert abs(manager.position[1] - checkxy[1]) in range(3)


def testManagerWindowCleanup(app):
    """
    Open migrate, clone, delete, newvm, details, host windows, close the
    connection, make sure they all disappear
    """
    manager = app.topwin
    manager.window_maximize()

    # Open delete window hitting a special code path, then close it
    manager.find("test-many-devices", "table cell").click()
    manager.find("Edit", "menu").click()
    manager.find("Delete", "menu item").click()
    delete = app.root.find_fuzzy("Delete", "frame")
    delete.find("storage-list").grab_focus()
    delete.window_close()

    # Open Clone window hitting a special code path, then close it
    manager.find("test-clone", "table cell").click()
    app.rawinput.pressKey("Menu")
    app.root.find("Clone...", "menu item").click()
    clone = app.find_window("Clone Virtual Machine")
    clone.window_close()

    # Open host
    manager.grab_focus()
    c = manager.find_fuzzy("testdriver.xml", "table cell")
    c.doubleClick()
    host = app.find_window("test testdriver.xml - Connection Details")

    # Open details
    manager.grab_focus()
    c = manager.find("test-many-devices", "table cell")
    c.doubleClick()
    details = app.find_details_window("test-many-devices")

    # Close the connection
    manager.grab_focus()
    app.manager_conn_disconnect("test testdriver.xml")

    # Ensure all those windows aren't showing
    lib.utils.check(lambda: not details.showing)

    # Delete the connection, ensure the host dialog disappears
    app.manager_conn_delete("test testdriver.xml")
    lib.utils.check(lambda: not host.showing)


def testManagerDefaultStartup(app):
    app.open(use_uri=False)
    manager = app.topwin
    errlabel = manager.find("error-label")
    lib.utils.check(lambda: "File->Add Connection" in errlabel.text, timeout=5)
    lib.utils.check(lambda: "appropriate QEMU/KVM" in errlabel.text)

    manager.find("File", "menu").click()
    manager.find("Quit", "menu item").click()


def testManagerConnOpenFail(app):
    app.open(keyfile="baduri.ini")
    manager = app.topwin
    manager.find_fuzzy("bad uri", "table cell").doubleClick()
    lib.utils.check(lambda: not manager.active)
    app.click_alert_button("Unable to connect", "Close")
    lib.utils.check(lambda: manager.active)



#
# Copyright 2010, 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceController(Device):
    XML_NAME = "controller"

    TYPE_IDE = "ide"
    TYPE_FDC = "fdc"
    TYPE_SCSI = "scsi"
    TYPE_SATA = "sata"
    TYPE_VIRTIOSERIAL = "virtio-serial"
    TYPE_USB = "usb"
    TYPE_PCI = "pci"
    TYPE_CCID = "ccid"
    TYPE_XENBUS = "xenbus"

    @staticmethod
    def get_usb2_controllers(conn):
        ret = []
        ctrl = DeviceController(conn)
        ctrl.type = "usb"
        ctrl.model = "ich9-ehci1"
        ret.append(ctrl)

        ctrl = DeviceController(conn)
        ctrl.type = "usb"
        ctrl.model = "ich9-uhci1"
        ctrl.master_startport = 0
        ret.append(ctrl)

        ctrl = DeviceController(conn)
        ctrl.type = "usb"
        ctrl.model = "ich9-uhci2"
        ctrl.master_startport = 2
        ret.append(ctrl)

        ctrl = DeviceController(conn)
        ctrl.type = "usb"
        ctrl.model = "ich9-uhci3"
        ctrl.master_startport = 4
        ret.append(ctrl)
        return ret

    @staticmethod
    def get_usb3_controller(conn, guest):
        ignore = guest
        ctrl = DeviceController(conn)
        ctrl.type = "usb"
        ctrl.model = "nec-xhci"
        if conn.support.conn_qemu_xhci():
            ctrl.model = "qemu-xhci"
        if conn.support.conn_usb3_ports():
            # 15 is the max ports qemu supports, might as well
            # Add as many as possible
            ctrl.ports = 15
        return ctrl

    _XML_PROP_ORDER = [
        "type",
        "index",
        "model",
        "master_startport",
        "driver_queues",
        "maxGrantFrames",
    ]

    type = XMLProperty("./@type")
    model = XMLProperty("./@model")
    vectors = XMLProperty("./@vectors", is_int=True)
    ports = XMLProperty("./@ports", is_int=True)
    maxGrantFrames = XMLProperty("./@maxGrantFrames", is_int=True)
    index = XMLProperty("./@index", is_int=True)

    driver_iothread = XMLProperty("./driver/@iothread", is_int=True)
    driver_queues = XMLProperty("./driver/@queues", is_int=True)

    master_startport = XMLProperty("./master/@startport", is_int=True)

    target_chassisNr = XMLProperty("./target/@chassisNr", is_int=True)
    target_chassis = XMLProperty("./target/@chassis", is_int=True)
    target_port = XMLProperty("./target/@port", is_int=True)
    target_hotplug = XMLProperty("./target/@hotplug", is_onoff=True)
    target_busNr = XMLProperty("./target/@busNr", is_int=True)
    target_index = XMLProperty("./target/@index", is_int=True)
    target_node = XMLProperty("./target/node", is_int=True)
    target_memReserve = XMLProperty("./target/@memReserve", is_int=True)

    def _get_attached_disk_devices(self, guest):
        ret = []
        for disk in guest.devices.disk:
            if self.type == disk.bus and self.index == disk.address.controller:
                ret.append(disk)
        return ret

    def _get_attached_virtioserial_devices(self, guest):
        ret = []
        for dev in guest.devices.channel:
            if self.type == dev.address.type and self.index == dev.address.controller:
                ret.append(dev)
        for dev in guest.devices.console:
            # virtio console is implied to be on virtio-serial index=0
            if self.index == 0 and dev.target_type == "virtio":
                ret.append(dev)
        return ret

    def get_attached_devices(self, guest):
        """
        Return all the Device objects from the passed Guest that are attached
        to this controller
        """
        ret = []
        if self.type == "virtio-serial":
            ret = self._get_attached_virtioserial_devices(guest)
        elif self.type in ["scsi", "sata", "ide", "fdc"]:
            ret = self._get_attached_disk_devices(guest)
        return ret



# Copyright (C) 2018 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re
import time

import libvirt

from virtinst import log

from ..baseclass import vmmGObject


class _VMStatsRecord:
    """
    Tracks a set of VM stats for a single timestamp
    """

    def __init__(
        self,
        timestamp,
        cpuTime,
        cpuTimeAbs,
        cpuHostPercent,
        cpuGuestPercent,
        curmem,
        currMemPercent,
        diskRdBytes,
        diskWrBytes,
        netRxBytes,
        netTxBytes,
    ):
        self.timestamp = timestamp
        self.cpuTime = cpuTime
        self.cpuTimeAbs = cpuTimeAbs
        self.cpuHostPercent = cpuHostPercent
        self.cpuGuestPercent = cpuGuestPercent
        self.curmem = curmem
        self.currMemPercent = currMemPercent
        self.diskRdKiB = diskRdBytes // 1024
        self.diskWrKiB = diskWrBytes // 1024
        self.netRxKiB = netRxBytes // 1024
        self.netTxKiB = netTxBytes // 1024

        # These are set in _VMStatsList.append_stats
        self.diskRdRate = None
        self.diskWrRate = None
        self.netRxRate = None
        self.netTxRate = None


class _VMStatsList(vmmGObject):
    """
    Tracks a list of VMStatsRecords for a single VM
    """

    def __init__(self):
        vmmGObject.__init__(self)
        self._stats = []

        self.diskRdMaxRate = 10.0
        self.diskWrMaxRate = 10.0
        self.netRxMaxRate = 10.0
        self.netTxMaxRate = 10.0

        self.mem_stats_period_is_set = False
        self.stats_disk_skip = []
        self.stats_net_skip = []

    def _cleanup(self):
        pass

    def append_stats(self, newstats):
        expected = self.config.get_stats_history_length()
        current = len(self._stats)
        if current > expected:  # pragma: no cover
            del self._stats[expected:current]

        def _calculate_rate(record_name):
            ret = 0.0
            if self._stats:
                oldstats = self._stats[0]
                ratediff = getattr(newstats, record_name) - getattr(oldstats, record_name)
                timediff = newstats.timestamp - oldstats.timestamp
                ret = float(ratediff) / float(timediff)
            return max(ret, 0.0)

        newstats.diskRdRate = _calculate_rate("diskRdKiB")
        newstats.diskWrRate = _calculate_rate("diskWrKiB")
        newstats.netRxRate = _calculate_rate("netRxKiB")
        newstats.netTxRate = _calculate_rate("netTxKiB")

        self.diskRdMaxRate = max(newstats.diskRdRate, self.diskRdMaxRate)
        self.diskWrMaxRate = max(newstats.diskWrRate, self.diskWrMaxRate)
        self.netRxMaxRate = max(newstats.netRxRate, self.netRxMaxRate)
        self.netTxMaxRate = max(newstats.netTxRate, self.netTxMaxRate)

        self._stats.insert(0, newstats)

    def get_record(self, record_name):
        if not self._stats:
            return 0
        return getattr(self._stats[0], record_name)

    def get_vector(self, record_name, limit, ceil=100.0):
        vector = []
        statslen = self.config.get_stats_history_length() + 1
        if limit is not None:
            statslen = min(statslen, limit)

        for i in range(statslen):
            if i < len(self._stats):
                vector.append(getattr(self._stats[i], record_name) / ceil)
            else:
                vector.append(0)
        return vector

    def get_in_out_vector(self, name1, name2, limit, ceil):
        return (self.get_vector(name1, limit, ceil=ceil), self.get_vector(name2, limit, ceil=ceil))


class vmmStatsManager(vmmGObject):
    """
    Class for polling statistics
    """

    def __init__(self):
        vmmGObject.__init__(self)
        self._vm_stats = {}
        self._latest_all_stats = {}

        self._all_stats_supported = True
        self._net_stats_supported = True
        self._disk_stats_supported = True
        self._disk_stats_lxc_supported = True
        self._mem_stats_supported = True

    def _cleanup(self):
        for statslist in self._vm_stats.values():
            statslist.cleanup()
        self._latest_all_stats = None

    ######################
    # CPU stats handling #
    ######################

    def _old_cpu_stats_helper(self, vm):
        info = vm.get_backend().info()
        state = info[0]
        guestcpus = info[3]
        cpuTimeAbs = info[4]
        return state, guestcpus, cpuTimeAbs

    def _sample_cpu_stats(self, vm, allstats):
        timestamp = time.time()
        if not vm.is_active() or not self.config.get_stats_enable_cpu_poll():
            return 0, 0, 0, 0, timestamp

        cpuTime = 0
        cpuHostPercent = 0
        cpuGuestPercent = 0
        prevTimestamp = self.get_vm_statslist(vm).get_record("timestamp")
        prevCpuTime = self.get_vm_statslist(vm).get_record("cpuTimeAbs")

        if allstats:
            state = allstats.get("state.state", 0)
            guestcpus = allstats.get("vcpu.current", 0)
            cpuTimeAbs = allstats.get("cpu.time", 0)
            timestamp = allstats.get("virt-manager.timestamp")
        else:
            state, guestcpus, cpuTimeAbs = self._old_cpu_stats_helper(vm)

        is_offline = state in [libvirt.VIR_DOMAIN_SHUTOFF, libvirt.VIR_DOMAIN_CRASHED]
        if is_offline:
            guestcpus = 0
            cpuTimeAbs = 0

        cpuTime = cpuTimeAbs - prevCpuTime
        if not is_offline:
            hostcpus = vm.conn.host_active_processor_count()

            pcentbase = ((cpuTime) * 100.0) / (
                (timestamp - prevTimestamp) * 1000.0 * 1000.0 * 1000.0
            )
            cpuHostPercent = pcentbase / hostcpus
            # Under RHEL-5.9 using a XEN HV guestcpus can be 0 during shutdown
            # so play safe and check it.
            cpuGuestPercent = guestcpus > 0 and pcentbase / guestcpus or 0

        cpuHostPercent = max(0.0, min(100.0, cpuHostPercent))
        cpuGuestPercent = max(0.0, min(100.0, cpuGuestPercent))

        return cpuTime, cpuTimeAbs, cpuHostPercent, cpuGuestPercent, timestamp

    ######################
    # net stats handling #
    ######################

    def _old_net_stats_helper(self, vm, dev):
        statslist = self.get_vm_statslist(vm)
        try:
            io = vm.get_backend().interfaceStats(dev)
            if io:
                rx = io[0]
                tx = io[4]
                return rx, tx
        except libvirt.libvirtError as err:  # pragma: no cover
            if vm.conn.support.is_error_nosupport(err):
                log.debug("conn does not support interfaceStats")
                self._net_stats_supported = False
                return 0, 0

            log.debug("Error in interfaceStats for '%s' dev '%s': %s", vm.get_name(), dev, err)
            if vm.is_active():
                log.debug("Adding %s to skip list", dev)
                statslist.stats_net_skip.append(dev)
            else:
                log.debug("Aren't running, don't add to skiplist")

        return 0, 0  # pragma: no cover

    def _sample_net_stats(self, vm, allstats):
        rx = 0
        tx = 0
        statslist = self.get_vm_statslist(vm)
        if (
            not self._net_stats_supported
            or not vm.is_active()
            or not self.config.get_stats_enable_net_poll()
        ):
            statslist.stats_net_skip = []
            return rx, tx

        if allstats:
            for key in allstats.keys():  # pragma: no cover
                if re.match(r"net.[0-9]+.rx.bytes", key):
                    rx += allstats[key]
                if re.match(r"net.[0-9]+.tx.bytes", key):
                    tx += allstats[key]
            return rx, tx

        for iface in vm.get_interface_devices_norefresh():
            dev = iface.target_dev
            if not dev:
                continue  # pragma: no cover
            if dev in statslist.stats_net_skip:
                continue  # pragma: no cover

            devrx, devtx = self._old_net_stats_helper(vm, dev)
            rx += devrx
            tx += devtx

        return rx, tx

    #######################
    # disk stats handling #
    #######################

    def _old_disk_stats_helper(self, vm, dev):
        statslist = self.get_vm_statslist(vm)
        try:
            io = vm.get_backend().blockStats(dev)
            if io:
                rd = io[1]
                wr = io[3]
                return rd, wr
        except libvirt.libvirtError as err:  # pragma: no cover
            if vm.conn.support.is_error_nosupport(err):
                log.debug("conn does not support blockStats")
                self._disk_stats_supported = False
                return 0, 0

            log.debug("Error in blockStats for '%s' dev '%s': %s", vm.get_name(), dev, err)
            if vm.is_active():
                log.debug("Adding %s to skip list", dev)
                statslist.stats_disk_skip.append(dev)
            else:
                log.debug("Aren't running, don't add to skiplist")

        return 0, 0  # pragma: no cover

    def _sample_disk_stats(self, vm, allstats):
        rd = 0
        wr = 0
        statslist = self.get_vm_statslist(vm)
        if (
            not self._disk_stats_supported
            or not vm.is_active()
            or not self.config.get_stats_enable_disk_poll()
        ):
            statslist.stats_disk_skip = []
            return rd, wr

        if allstats:
            for key in allstats.keys():
                if re.match(r"block.[0-9]+.rd.bytes", key):
                    rd += allstats[key]
                if re.match(r"block.[0-9]+.wr.bytes", key):
                    wr += allstats[key]
            return rd, wr

        # LXC has a special blockStats method
        if vm.conn.is_lxc() and self._disk_stats_lxc_supported:
            try:
                io = vm.get_backend().blockStats("")
                if io:
                    rd = io[1]
                    wr = io[3]
                    return rd, wr
            except libvirt.libvirtError as e:  # pragma: no cover
                log.debug("LXC style disk stats not supported: %s", e)
                self._disk_stats_lxc_supported = False

        for disk in vm.get_disk_devices_norefresh():
            dev = disk.target
            if not dev:
                continue  # pragma: no cover
            if dev in statslist.stats_disk_skip:
                continue  # pragma: no cover

            diskrd, diskwr = self._old_disk_stats_helper(vm, dev)
            rd += diskrd
            wr += diskwr

        return rd, wr

    #########################
    # memory stats handling #
    #########################

    def _set_mem_stats_period(self, vm):
        # QEMU requires to explicitly enable memory stats polling per VM
        # if we want fine grained memory stats
        if not vm.conn.support.conn_mem_stats_period():
            return

        # Only works for virtio balloon
        if not any([b for b in vm.get_xmlobj().devices.memballoon if b.model == "virtio"]):
            return  # pragma: no cover

        try:
            secs = 5
            vm.get_backend().setMemoryStatsPeriod(secs, libvirt.VIR_DOMAIN_AFFECT_LIVE)
        except Exception as e:  # pragma: no cover
            log.debug("Error setting memstats period: %s", e)

    def _old_mem_stats_helper(self, vm):
        totalmem = 1
        curmem = 0
        try:
            stats = vm.get_backend().memoryStats()
            totalmem = stats.get("actual", 1)
            curmem = max(0, totalmem - stats.get("unused", totalmem))
        except libvirt.libvirtError as err:  # pragma: no cover
            if vm.conn.support.is_error_nosupport(err):
                log.debug("conn does not support memoryStats")
                self._mem_stats_supported = False
            else:
                log.debug("Error reading mem stats: %s", err)

        return totalmem, curmem

    def _sample_mem_stats(self, vm, allstats):
        statslist = self.get_vm_statslist(vm)
        if (
            not self._mem_stats_supported
            or not vm.is_active()
            or not self.config.get_stats_enable_memory_poll()
        ):
            statslist.mem_stats_period_is_set = False
            return 0, 0

        if statslist.mem_stats_period_is_set is False:
            self._set_mem_stats_period(vm)
            statslist.mem_stats_period_is_set = True

        if allstats:
            totalmem = allstats.get("balloon.current", 1)
            curmem = max(0, totalmem - allstats.get("balloon.unused", totalmem))
        else:
            totalmem, curmem = self._old_mem_stats_helper(vm)

        currMemPercent = (curmem / float(totalmem)) * 100
        currMemPercent = max(0.0, min(currMemPercent, 100.0))

        return currMemPercent, curmem

    ####################
    # alltats handling #
    ####################

    def _get_all_stats(self, conn):
        # test conn supports allstats as of 2021, but for test coverage
        # purposes lets still use the old stats code for the test driver
        if not self._all_stats_supported or conn.is_test():
            return {}

        statflags = 0
        if self.config.get_stats_enable_cpu_poll():
            statflags |= libvirt.VIR_DOMAIN_STATS_STATE
            statflags |= libvirt.VIR_DOMAIN_STATS_CPU_TOTAL
            statflags |= libvirt.VIR_DOMAIN_STATS_VCPU
        if self.config.get_stats_enable_memory_poll():
            statflags |= libvirt.VIR_DOMAIN_STATS_BALLOON
        if self.config.get_stats_enable_disk_poll():
            statflags |= libvirt.VIR_DOMAIN_STATS_BLOCK
        if self.config.get_stats_enable_net_poll():
            statflags |= libvirt.VIR_DOMAIN_STATS_INTERFACE
        if statflags == 0:
            return {}  # pragma: no cover

        ret = {}
        try:
            timestamp = time.time()
            rawallstats = conn.get_backend().getAllDomainStats(statflags, 0)

            # Reformat the output to be a bit more friendly
            for dom, domallstats in rawallstats:
                domallstats["virt-manager.timestamp"] = timestamp
                ret[dom.UUIDString()] = domallstats
        except libvirt.libvirtError as err:
            if conn.support.is_error_nosupport(err):
                log.debug("conn does not support getAllDomainStats()")
                self._all_stats_supported = False
            else:  # pragma: no cover
                log.debug("Error call getAllDomainStats(): %s", err)
        return ret

    ##############
    # Public API #
    ##############

    def refresh_vm_stats(self, vm):
        domallstats = self._latest_all_stats.get(vm.get_uuid(), None)

        (cpuTime, cpuTimeAbs, cpuHostPercent, cpuGuestPercent, timestamp) = self._sample_cpu_stats(
            vm, domallstats
        )
        currMemPercent, curmem = self._sample_mem_stats(vm, domallstats)
        diskRdBytes, diskWrBytes = self._sample_disk_stats(vm, domallstats)
        netRxBytes, netTxBytes = self._sample_net_stats(vm, domallstats)

        newstats = _VMStatsRecord(
            timestamp,
            cpuTime,
            cpuTimeAbs,
            cpuHostPercent,
            cpuGuestPercent,
            curmem,
            currMemPercent,
            diskRdBytes,
            diskWrBytes,
            netRxBytes,
            netTxBytes,
        )
        self.get_vm_statslist(vm).append_stats(newstats)

    def cache_all_stats(self, conn):
        self._latest_all_stats = self._get_all_stats(conn)

    def get_vm_statslist(self, vm):
        if vm.get_name() not in self._vm_stats:
            self._vm_stats[vm.get_name()] = _VMStatsList()
        return self._vm_stats[vm.get_name()]



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


#################################################
# UI tests for virt-manager's VM details window #
#################################################


def testHostNetworkSmokeTest(app):
    """
    Verify that each virtual network displays, without error.
    """
    win = app.manager_open_host("Virtual Networks")
    lst = win.find("net-list", "table")
    errlabel = win.find("net-error-label", "label")
    lib.utils.walkUIList(app, win, lst, lambda: errlabel.showing)

    # Select XML editor, and reverse walk the list
    win.find("network-grid").find("XML", "page tab").click()
    lib.utils.walkUIList(app, win, lst, lambda: errlabel.showing, reverse=True)


def testHostNetworkEdit(app):
    """
    Test edits to net config
    """
    app.open(xmleditor_enabled=True)
    win = app.manager_open_host("Virtual Networks").find("network-grid")
    finish = win.find("Apply", "push button")

    # Shut it off, do an XML edit, verify it
    win.find("default", "table cell").click()
    delete = win.find("net-delete", "push button")
    stop = win.find("net-stop", "push button")
    stop.click()
    lib.utils.check(lambda: delete.sensitive)
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    origdev = "virbr0"
    newdev = "virbr77"
    xmleditor.set_text(xmleditor.text.replace(origdev, newdev))
    finish.click()
    win.find("Details", "page tab").click()
    netdev = win.find("net-device")
    lib.utils.check(lambda: netdev.text == newdev)

    # Rename it
    win.find("default", "table cell").click()
    win.find("net-name").set_text("newsort-default")
    finish.click()

    # Change autostart, trigger it by clicking away
    win.find("newsort-default", "table cell").click()
    win.find("net-autostart").click()
    win.find("netboot", "table cell").click()
    app.click_alert_button("There are unapplied changes", "Yes")

    # Do standard xmleditor tests
    lib.utils.test_xmleditor_interactions(app, win, finish)


def testHostStorageSmokeTest(app):
    """
    Verify that each storage pool displays, without error.
    """
    win = app.manager_open_host("Storage")
    lst = win.find("pool-list", "table")
    errlabel = win.find("pool-error-label", "label")
    lib.utils.walkUIList(app, win, lst, lambda: errlabel.showing)

    # Select XML editor, and reverse walk the list
    win.find("storage-grid").find("XML", "page tab").click()
    lib.utils.walkUIList(app, win, lst, lambda: errlabel.showing, reverse=True)


def testHostStorageEdit(app):
    """
    Test edits to pool config
    """
    app.open(xmleditor_enabled=True)
    win = app.manager_open_host("Storage").find("storage-grid")
    finish = win.find("Apply", "push button")

    # Shut off a pool, do an XML edit, verify it
    win.find("pool-dir", "table cell").click()
    delete = win.find("pool-delete", "push button")
    stop = win.find("pool-stop", "push button")
    stop.click()
    lib.utils.check(lambda: delete.sensitive)
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    origpath = "/pool-dir"
    newpath = "/dev/foo/bar/baz"
    xmleditor.set_text(xmleditor.text.replace(origpath, newpath))
    finish.click()
    win.find("Details", "page tab").click()
    poolloc = win.find("pool-location")
    lib.utils.check(lambda: poolloc.text == newpath)

    # Rename it
    win.find("pool-dir", "table cell").click()
    win.find("pool-name").set_text("newsort-default")
    finish.click()

    # Change autostart. Trigger it by clicking on new cell
    win.find("newsort-default", "table cell").click()
    win.find("pool-autostart").click()
    win.find("pool-logical", "table cell").click()
    app.click_alert_button("There are unapplied changes", "Yes")

    # Do standard xmleditor tests
    lib.utils.test_xmleditor_interactions(app, win, finish)


def testHostStorageVolMisc(app):
    """
    Misc actions involving volumes
    """
    win = app.manager_open_host("Storage").find("storage-grid")
    win.find_fuzzy("pool-dir", "table cell").click()
    vollist = win.find("vol-list", "table")

    vol1 = vollist.find("backingl1.img", "table cell")
    vol2 = vollist.find("UPPER", "table cell")
    vol1.check_onscreen()
    vol2.check_not_onscreen()
    win.find("Size", "table column header").click()
    win.find("Size", "table column header").click()
    vol1.check_not_onscreen()
    vol2.check_onscreen()

    vol2.click(button=3)
    app.root.find("Copy Volume Path", "menu item").click()
    from gi.repository import Gdk, Gtk

    clipboard = Gtk.Clipboard.get_default(Gdk.Display.get_default())
    lib.utils.check(lambda: clipboard.wait_for_text() == "/pool-dir/UPPER")


def testHostConn(app):
    """
    Change some connection parameters
    """
    manager = app.topwin
    # Disconnect the connection
    app.manager_conn_disconnect("test testdriver.xml")

    # Open Host Details from right click menu
    c = manager.find("test testdriver.xml", "table cell")
    c.click(button=3)
    app.root.find("conn-details", "menu item").click()
    win = app.find_window("test testdriver.xml - Connection Details")

    # Click the tabs and then back
    win.find_fuzzy("Storage", "tab").click()
    win.find_fuzzy("Network", "tab").click()
    win.find_fuzzy("Overview", "tab").click()

    # Toggle autoconnect
    win.find("Autoconnect:", "check box").click()
    win.find("Autoconnect:", "check box").click()

    # Change the name, verify that title bar changed
    win.find("Name:", "text").set_text("FOOBAR")
    app.find_window("FOOBAR - Connection Details")

    # Open the manager window
    win.find("File", "menu").click()
    win.find("View Manager", "menu item").click()
    lib.utils.check(lambda: manager.active)
    # Confirm connection row is named differently in manager
    manager.find("FOOBAR", "table cell")

    # Close the manager
    manager.window_close()
    lib.utils.check(lambda: win.active)

    # Quit app from the file menu
    win.find("File", "menu").click()
    win.find("Quit", "menu item").click()
    lib.utils.check(lambda: not app.is_running())



# Copyright (C) 2009, 2013 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import gi
from gi.repository import Gio
from gi.repository import Gtk

from virtinst import log
from virtinst import xmlutil

from . import vmmenu
from .baseclass import vmmGObject
from .connmanager import vmmConnectionManager


# pylint: disable=ungrouped-imports
try:  # pragma: no cover
    # pylint: disable=no-name-in-module
    gi.require_version("AyatanaAppIndicator3", "0.1")
    from gi.repository import AyatanaAppIndicator3 as AppIndicator3
except Exception:  # pragma: no cover
    AppIndicator3 = None


def _toggle_manager(*args, **kwargs):
    ignore = args
    ignore = kwargs
    from .manager import vmmManager

    manager = vmmManager.get_instance(None)
    if manager.is_visible():
        manager.close()
    else:
        manager.show()


def _conn_connect_cb(src, uri):
    connmanager = vmmConnectionManager.get_instance()
    conn = connmanager.conns[uri]
    if conn.is_disconnected():
        conn.open()


def _conn_disconnect_cb(src, uri):
    connmanager = vmmConnectionManager.get_instance()
    conn = connmanager.conns[uri]
    if not conn.is_disconnected():
        conn.close()


def _has_appindicator_dbus():  # pragma: no cover
    try:
        bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        dbus = Gio.DBusProxy.new_sync(
            bus,
            0,
            None,
            "org.freedesktop.DBus",
            "/org/freedesktop/DBus",
            "org.freedesktop.DBus",
            None,
        )
        if dbus.NameHasOwner("(s)", "org.kde.StatusNotifierWatcher"):
            return True
        if dbus.NameHasOwner("(s)", "org.freedesktop.StatusNotifierWatcher"):
            return True
        return False
    except Exception:
        log.exception("Error checking for appindicator dbus")
        return False


_USING_APPINDICATOR = False
if AppIndicator3:  # pragma: no cover
    log.debug("Imported AppIndicator3=%s", AppIndicator3)
    if not _has_appindicator_dbus():
        log.debug("AppIndicator3 is available, but didn't find any dbus watcher.")
    else:
        _USING_APPINDICATOR = True
        log.debug("Using AppIndicator3 for systray")


###########################
# systray backend classes #
###########################


class _Systray:
    def is_embedded(self):
        raise NotImplementedError()

    def show(self):
        raise NotImplementedError()

    def hide(self):
        raise NotImplementedError()

    def set_menu(self, menu):
        raise NotImplementedError()


class _SystrayIndicator(_Systray):  # pragma: no cover
    """
    UI backend for appindicator
    """

    def __init__(self):
        self._icon = AppIndicator3.Indicator.new(
            "virt-manager", "virt-manager", AppIndicator3.IndicatorCategory.APPLICATION_STATUS
        )

    def set_menu(self, menu):
        hide_item = Gtk.MenuItem.new_with_mnemonic(_("_Show Virtual Machine Manager"))
        hide_item.connect("activate", _toggle_manager)
        hide_item.show()
        menu.insert(hide_item, len(menu.get_children()) - 1)

        self._icon.set_menu(menu)
        self._icon.set_secondary_activate_target(hide_item)

    def is_embedded(self):
        if not self._icon.get_property("connected"):
            return False
        return self._icon.get_status() != AppIndicator3.IndicatorStatus.PASSIVE

    def show(self):
        self._icon.set_status(AppIndicator3.IndicatorStatus.ACTIVE)

    def hide(self):
        self._icon.set_status(AppIndicator3.IndicatorStatus.PASSIVE)


class _SystrayStatusIcon(_Systray):  # pragma: no cover
    """
    UI backend for Gtk StatusIcon
    """

    def __init__(self):
        self._icon = Gtk.StatusIcon()
        self._icon.set_property("icon-name", "virt-manager")
        self._icon.connect("activate", _toggle_manager)
        self._icon.connect("popup-menu", self._popup_cb)
        self._icon.set_tooltip_text(_("Virtual Machine Manager"))
        self._menu = None

    def is_embedded(self):
        return self._icon.is_embedded()

    def set_menu(self, menu):
        self._menu = menu

    def _popup_cb(self, src, button, event_time):
        if button != 3:
            return

        self._menu.popup(None, None, Gtk.StatusIcon.position_menu, self._icon, 0, event_time)

    def show(self):
        self._icon.set_visible(True)

    def hide(self):
        self._icon.set_visible(False)


class _SystrayWindow(_Systray):
    """
    A mock systray implementation that shows its own top level window,
    so we can test more of the infrastructure in our ui tests
    """

    def __init__(self):
        self._window = None
        self._menu = None
        self._init_ui()

    def _init_ui(self):
        button = Gtk.Button.new_from_stock(Gtk.STOCK_ADD)
        button.connect("button-press-event", self._popup_cb)

        self._window = Gtk.Window()
        self._window.set_size_request(100, 100)
        self._window.get_accessible().set_name("vmm-fake-systray")
        self._window.add(button)

    def is_embedded(self):
        return self._window.is_visible()

    def set_menu(self, menu):
        self._menu = menu

    def _popup_cb(self, src, event):
        if event.button == 1:
            _toggle_manager()
        else:
            self._menu.popup_at_pointer(event)

    def show(self):
        self._window.show_all()

    def hide(self):
        self._window.hide()


class _TrayMainMenu(vmmGObject):
    """
    Helper class for maintaining the conn + VM menu list and updating
    it in place
    """

    def __init__(self):
        vmmGObject.__init__(self)
        self.topwin = None  # Need this for error callbacks from VMActionMenu

        self._menu = self._build_menu()

    def _cleanup(self):
        self._menu.destroy()
        self._menu = None

    ###########
    # UI init #
    ###########

    def _build_menu(self):
        """
        Build the top level conn list menu when clicking the icon
        """
        menu = Gtk.Menu()
        menu.get_accessible().set_name("vmm-systray-menu")
        menu.add(Gtk.SeparatorMenuItem())

        exit_item = Gtk.ImageMenuItem.new_from_stock(Gtk.STOCK_QUIT, None)
        exit_item.connect("activate", self._exit_app_cb)
        menu.add(exit_item)
        menu.show_all()
        return menu

    ######################
    # UI update routines #
    ######################

    # Helpers for stashing identifying data in the menu item objects
    def _get_lookupkey(self, child):
        return getattr(child, "_vmlookupkey", None)

    def _set_lookupkey(self, child, val):
        return setattr(child, "_vmlookupkey", val)

    def _get_sortkey(self, child):
        return getattr(child, "_vmsortkey", None)

    def _set_sortkey(self, child, val):
        return setattr(child, "_vmsortkey", val)

    def _set_vm_state(self, menu_item, vm):
        label = menu_item.get_child()
        label.set_text(vm.get_name_or_title())
        vm_action_menu = menu_item.get_submenu()
        vm_action_menu.update_widget_states(vm)

    def _build_vm_menuitem(self, vm):
        """
        Build a menu item representing a single VM
        """
        menu_item = Gtk.ImageMenuItem.new_with_label("FOO")
        menu_item.set_use_underline(False)
        vm_action_menu = vmmenu.VMActionMenu(self, lambda: vm)
        menu_item.set_submenu(vm_action_menu)
        self._set_lookupkey(menu_item, vm)
        self._set_sortkey(menu_item, vm.get_name_or_title())
        self._set_vm_state(menu_item, vm)
        menu_item.show_all()
        return menu_item

    def _set_conn_state(self, menu_item, conn):
        label = menu_item.get_child()
        if conn.is_active():
            label = menu_item.get_child()
            markup = "<b>%s</b>" % xmlutil.xml_escape(conn.get_pretty_desc())
            label.set_markup(markup)
        else:
            label.set_text(conn.get_pretty_desc())

        connect_item = self._find_lookupkey(menu_item.get_submenu(), 1)
        disconnect_item = self._find_lookupkey(menu_item.get_submenu(), 2)
        connect_item.set_visible(conn.is_active())
        disconnect_item.set_visible(not conn.is_active())

    def _build_conn_menuitem(self, conn):
        """
        Build a menu item representing a single connection, and populate
        all its VMs as items in a sub menu
        """
        menu_item = Gtk.MenuItem.new_with_label("FOO")
        self._set_lookupkey(menu_item, conn.get_uri())

        # Group active conns first
        # Sort by pretty desc within those categories
        sortkey = str(int(bool(not conn.is_active())))
        sortkey += conn.get_pretty_desc().lower()
        self._set_sortkey(menu_item, sortkey)

        menu = Gtk.Menu()
        menu_item.set_submenu(menu)

        menu.add(Gtk.SeparatorMenuItem())
        citem1 = Gtk.ImageMenuItem.new_from_stock(Gtk.STOCK_DISCONNECT, None)
        citem1.connect("activate", _conn_disconnect_cb, conn.get_uri())
        self._set_lookupkey(citem1, 1)
        menu.add(citem1)
        citem2 = Gtk.ImageMenuItem.new_from_stock(Gtk.STOCK_CONNECT, None)
        citem2.connect("activate", _conn_connect_cb, conn.get_uri())
        self._set_lookupkey(citem2, 2)
        menu.add(citem2)

        menu_item.show_all()
        self._set_conn_state(menu_item, conn)
        return menu_item

    def _find_lookupkey(self, parent, key):
        for child in parent.get_children():
            if self._get_lookupkey(child) == key:
                return child

    def _find_conn_menuitem(self, uri):
        return self._find_lookupkey(self._menu, uri)

    def _find_vm_menuitem(self, uri, vm):
        connmenu = self._find_conn_menuitem(uri)
        return self._find_lookupkey(connmenu.get_submenu(), vm)

    ################
    # UI listeners #
    ################

    def _exit_app_cb(self, src):
        from .engine import vmmEngine

        vmmEngine.get_instance().exit_app()

    ##############
    # Public API #
    ##############

    def get_menu(self):
        return self._menu

    def conn_add(self, conn):
        connmenu = self._build_conn_menuitem(conn)
        sortkey = self._get_sortkey(connmenu)

        idx = 0
        for idx, child in enumerate(list(self._menu.get_children())):
            checksort = self._get_sortkey(child)
            if checksort is None or checksort > sortkey:
                break

        self._menu.insert(connmenu, idx)

    def conn_remove(self, uri):
        connmenu = self._find_conn_menuitem(uri)
        if connmenu:
            self._menu.remove(connmenu)
            connmenu.destroy()

    def conn_change(self, conn):
        connmenu = self._find_conn_menuitem(conn.get_uri())
        self._set_conn_state(connmenu, conn)

    def vm_add(self, vm):
        connmenu = self._find_conn_menuitem(vm.conn.get_uri())
        menu_item = self._build_vm_menuitem(vm)
        sortkey = self._get_sortkey(menu_item)

        vmsubmenu = connmenu.get_submenu()
        idx = 0
        for idx, child in enumerate(list(vmsubmenu.get_children())):
            checksort = self._get_sortkey(child)
            if checksort is None or checksort > sortkey:
                break

        vmsubmenu.insert(menu_item, idx)

    def vm_remove(self, vm):
        conn = vm.conn
        connmenu = self._find_conn_menuitem(conn.get_uri())
        vmitem = self._find_vm_menuitem(conn.get_uri(), vm)
        connmenu.get_submenu().remove(vmitem)
        vmitem.destroy()

    def vm_change(self, vm):
        vmitem = self._find_vm_menuitem(vm.conn.get_uri(), vm)
        self._set_vm_state(vmitem, vm)


class vmmSystray(vmmGObject):
    """
    API class representing a systray icon. May use StatusIcon or appindicator
    backends
    """

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            cls._instance = vmmSystray()
        return cls._instance

    @staticmethod
    def systray_disabled_message():  # pragma: no cover
        if "WAYLAND_DISPLAY" not in os.environ:
            return
        if _USING_APPINDICATOR:
            return
        return _("AppIndicator not installed or usable, system tray icon not available on Wayland")

    def __init__(self):
        vmmGObject.__init__(self)
        self._cleanup_on_app_close()

        self._systray = None
        self._mainmenu = None

        self.add_gsettings_handle(
            self.config.on_view_system_tray_changed(self._show_systray_changed_cb)
        )
        self._startup()

    def is_embedded(self):
        return self._systray and self._systray.is_embedded()

    def show_from_cli(self):
        self._show_systray()

    def _cleanup(self):
        self._hide_systray()
        self._systray = None
        if self._mainmenu:
            self._mainmenu.cleanup()
            self._mainmenu = None

    ###########################
    # Initialization routines #
    ###########################

    def _init_mainmenu(self):
        self._mainmenu = _TrayMainMenu()
        connmanager = vmmConnectionManager.get_instance()
        connmanager.connect("conn-added", self._conn_added_cb)
        connmanager.connect("conn-removed", self._conn_removed_cb)
        for conn in connmanager.conns.values():
            self._conn_added_cb(connmanager, conn)

    def _show_systray(self):
        if not self._systray:
            if self.config.CLITestOptions.fake_systray:
                self._systray = _SystrayWindow()
            elif _USING_APPINDICATOR:  # pragma: no cover
                self._systray = _SystrayIndicator()
            else:  # pragma: no cover
                self._systray = _SystrayStatusIcon()
            self._init_mainmenu()
            self._systray.set_menu(self._mainmenu.get_menu())
        self._systray.show()

    def _hide_systray(self):
        if not self._systray:
            return
        self._systray.hide()

    def _show_systray_changed_cb(self):
        do_show = self.config.get_view_system_tray()
        log.debug("Showing systray: %s", do_show)

        if do_show:
            self._show_systray()
        else:
            self._hide_systray()

    def _startup(self):
        # This will trigger the actual UI showing
        self._show_systray_changed_cb()

    ################
    # UI listeners #
    ################

    def _conn_added_cb(self, src, conn):
        conn.connect("vm-added", self._vm_added_cb)
        conn.connect("vm-removed", self._vm_removed_cb)
        conn.connect("state-changed", self._conn_state_changed_cb)
        self._mainmenu.conn_add(conn)
        for vm in conn.list_vms():
            self._vm_added_cb(conn, vm)

    def _conn_removed_cb(self, src, conn):
        self._mainmenu.conn_remove(conn)

    def _conn_state_changed_cb(self, conn):
        self._mainmenu.conn_change(conn)

    def _vm_added_cb(self, conn, vm):
        vm.connect("state-changed", self._vm_state_changed_cb)
        self._mainmenu.vm_add(vm)

    def _vm_removed_cb(self, conn, vm):
        self._mainmenu.vm_remove(vm)

    def _vm_state_changed_cb(self, vm):
        self._mainmenu.vm_change(vm)
